# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from abc import abstractmethod
from typing import Dict, Any, Optional, List, Sequence, TypeVar, Generic, Type

from xcube.core.store import DatasetDescriptor
from xcube.util.assertions import assert_in
from xcube.util.assertions import assert_instance
from xcube.util.jsonschema import JsonArraySchema
from xcube.util.jsonschema import JsonIntegerSchema
from xcube.util.jsonschema import JsonObject
from xcube.util.jsonschema import JsonObjectSchema
from xcube.util.jsonschema import JsonStringSchema

STATUS_IDS = ['ok', 'error', 'warning']

R = TypeVar('R', bound=JsonObject)


class GenericCubeGeneratorResult(Generic[R], JsonObject):
    def __init__(self,
                 status: str,
                 status_code: Optional[int] = None,
                 result: Optional[R] = None,
                 message: Optional[str] = None,
                 output: Optional[Sequence[str]] = None,
                 traceback: Optional[Sequence[str]] = None,
                 versions: Optional[Dict[str, str]] = None):
        assert_instance(status, str, name='status')
        assert_in(status, STATUS_IDS, name='status')
        self.status = status
        self.status_code = status_code
        self.result = result
        self.message = message if message else None
        self.output = list(output) if output else None
        self.traceback = list(traceback) if traceback else None
        self.versions = dict(versions) if versions else None

    def derive(self,
               /,
               status: Optional[str] = None,
               status_code: Optional[int] = None,
               result: Optional[R] = None,
               message: Optional[str] = None,
               output: Optional[Sequence[str]] = None,
               traceback: Optional[Sequence[str]] = None,
               versions: Optional[Dict[str, str]] = None) -> R:
        return self.__class__(status or self.status,
                              status_code=status_code or self.status_code,
                              result=result or self.result,
                              message=message or self.message,
                              output=output or self.output,
                              traceback=traceback or self.traceback,
                              versions=versions or self.versions)

    @classmethod
    @abstractmethod
    def get_result_schema(cls) -> JsonObjectSchema:
        """Get the JSON schema of the result object"""

    @classmethod
    def from_dict(cls, value: Dict) -> R:
        return cls.get_schema().from_instance(value)

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                status=JsonStringSchema(enum=STATUS_IDS),
                status_code=JsonIntegerSchema(),
                result=cls.get_result_schema(),
                message=JsonStringSchema(),
                output=JsonArraySchema(items=JsonStringSchema()),
                traceback=JsonArraySchema(items=JsonStringSchema()),
                versions=JsonObjectSchema(additional_properties=True)
            ),
            required=['status'],
            additional_properties=True,
            factory=cls,
        )


def make_cube_generator_result_class(result_type: Type[R]) \
        -> Type[GenericCubeGeneratorResult[R]]:
    class SpecificCubeGeneratorResult(GenericCubeGeneratorResult[R]):
        @classmethod
        def get_result_schema(cls) -> JsonObjectSchema:
            return result_type.get_schema()

    return SpecificCubeGeneratorResult


class CubeReference(JsonObject):
    def __init__(self, data_id: str):
        self.data_id = data_id

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                data_id=JsonStringSchema(min_length=1),
            ),
            required=['data_id'],
            additional_properties=False,
            factory=cls,
        )


CubeGeneratorResult = make_cube_generator_result_class(CubeReference)


class CubeInfo(JsonObject):
    def __init__(self,
                 dataset_descriptor: DatasetDescriptor,
                 size_estimation: Dict[str, Any],
                 **kwargs):
        self.dataset_descriptor: DatasetDescriptor = dataset_descriptor
        self.size_estimation: dict = size_estimation

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(
            properties=dict(
                dataset_descriptor=DatasetDescriptor.get_schema(),
                size_estimation=JsonObjectSchema(additional_properties=True)
            ),
            required=['dataset_descriptor', 'size_estimation'],
            additional_properties=True,
            factory=cls
        )


CubeInfoResult = make_cube_generator_result_class(CubeInfo)



# The MIT License (MIT)
# Copyright (c) 2020-2024 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import warnings
from typing import Iterator, Any, Tuple, Container, Union, Dict, List

import fsspec
import xarray as xr

from xcube.util.jsonschema import JsonObjectSchema

from ..datatype import DataTypeLike
from ..descriptor import DataDescriptor, DatasetDescriptor
from ..descriptor import new_data_descriptor
from ..store import DataStore
from .schema import REF_STORE_SCHEMA


class ReferenceDataStore(DataStore):
    """A data store that uses kerchunk reference files to
    open referred NetCDF datasets as if they were Zarr datasets.

    :param refs: The list of reference JSON files to use for this
        instance. Files (URLs or local paths) are used in
        conjunction with target_options and target_protocol
        to open and parse JSON at this location.
    :param target_protocol: Used for loading the reference files.
        If not given, protocol will be derived from the given path.
    :param target_options: Extra filesystem options for loading
        the reference files.
    :param remote_protocol: The protocol of the filesystem on which
        the references will be evaluated. If not given, will be derived
        from the first URL in the references that has a protocol.
    :param remote_options: Extra filesystem options for loading the
        referenced data.
    :param max_gap: See ``max_block``.
    :param max_block: For merging multiple concurrent requests to the same
        remote file. Neighboring byte ranges will only be
        merged when their inter-range gap is <= `max_gap`.
        Default is 64KB. Set to 0 to only merge when it
        requires no extra bytes. Pass a negative number to
        disable merging, appropriate for local target files.
        Neighboring byte ranges will only be merged when the
        size of the aggregated range is <= ``max_block``.
        Default is 256MB.
    :param cache_size: Maximum size of LRU cache, where
        cache_size*record_size denotes the total number of
        references that can be loaded in memory at once.
        Only used for lazily loaded references.
    """

    def __init__(
            self,
            refs: List[str],
            **ref_kwargs,
    ):
        self._refs = dict(self._normalize_ref(ref) for ref in refs)
        self._ref_kwargs = ref_kwargs

    @classmethod
    def get_data_store_params_schema(cls) -> JsonObjectSchema:
        return REF_STORE_SCHEMA

    @classmethod
    def get_data_types(cls) -> Tuple[str, ...]:
        return ("dataset",)

    def get_data_types_for_data(self, data_id: str) -> Tuple[str, ...]:
        return self.get_data_types()

    def get_data_ids(self,
                     data_type: DataTypeLike = None,
                     include_attrs: Container[str] = None) -> \
            Union[Iterator[str], Iterator[Tuple[str, Dict[str, Any]]]]:
        return iter(self._refs.keys())

    def has_data(self,
                 data_id: str,
                 data_type: DataTypeLike = None) -> bool:
        return data_id in self._refs

    def describe_data(self,
                      data_id: str,
                      data_type: DataTypeLike = None) -> DataDescriptor:
        data_descriptor = self._refs[data_id].get("data_descriptor")
        if data_descriptor is None:
            dataset = self.open_data(data_id)
            data_descriptor = new_data_descriptor(data_id, dataset)
            self._refs[data_id]["data_descriptor"] = data_descriptor
        return data_descriptor

    def get_data_opener_ids(self,
                            data_id: str = None,
                            data_type: DataTypeLike = None) -> Tuple[str, ...]:
        return ("dataset:zarr:reference",)

    def get_open_data_params_schema(self,
                                    data_id: str = None,
                                    opener_id: str = None) -> JsonObjectSchema:
        # We do not have open parameters yet
        return JsonObjectSchema()

    def open_data(self,
                  data_id: str,
                  opener_id: str = None,
                  **open_params) -> xr.Dataset:
        if open_params:
            warnings.warn(f"open_params are not supported yet,"
                          f" but passing forward {', '.join(open_params.keys())}")
        ref_path = self._refs[data_id]["ref_path"]
        open_params.pop("consolidated", False)
        ref_mapping = fsspec.get_mapper('reference://', fo=ref_path, **self._ref_kwargs)
        return xr.open_zarr(ref_mapping, consolidated=False, **open_params)

    @classmethod
    def get_search_params_schema(cls,
                                 data_type: DataTypeLike = None) -> JsonObjectSchema:
        # We do not have search parameters yet
        return JsonObjectSchema()

    def search_data(self,
                    data_type: DataTypeLike = None,
                    **search_params) -> Iterator[DataDescriptor]:
        if search_params:
            warnings.warn(f"search_params are not supported yet,"
                          f" but received {', '.join(search_params.keys())}")
        return (self.describe_data(data_id) for data_id in self.get_data_ids())

    @classmethod
    def _normalize_ref(cls, ref: Union[str, Dict[str, Any]]) \
            -> Tuple[str, Dict[str, Any]]:

        if isinstance(ref, str):
            ref_path = ref
            data_id = cls._ref_path_to_data_id(ref_path)
            return data_id, dict(ref_path=ref_path, data_descriptor=None)

        if isinstance(ref, dict):
            ref_path = ref.get("ref_path") or None
            if not ref_path:
                raise ValueError("missing key ref_path in refs item")

            data_descriptor = ref.get("data_descriptor") or None
            if isinstance(data_descriptor, dict):
                data_descriptor = DatasetDescriptor.from_dict(data_descriptor)
            elif data_descriptor is not None:
                raise TypeError("value of data_descriptor key"
                                " in refs item must be a dict or None")

            data_id = ref.get("data_id") or None
            if data_id is None and data_descriptor is not None:
                data_id = data_descriptor.data_id
            if data_id is None:
                data_id = cls._ref_path_to_data_id(ref_path)

            return data_id, dict(ref_path=ref_path,
                                 data_descriptor=data_descriptor)
        raise TypeError("item in refs must be a str or a dict")

    @classmethod
    def _ref_path_to_data_id(cls, ref_path: str) -> str:
        protocol, path = fsspec.core.split_protocol(ref_path)
        if protocol in (None, "file", "local") and os.path.sep != "/":
            path = path.replace(os.path.sep, "/")
        name = path.rsplit("/", maxsplit=1)[-1]
        return name[:-5] if name.endswith(".json") else name



# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import fnmatch
import os.path
import re
import zipfile
from typing import Any, Dict, Optional, List, Collection, Iterator, Union

import fsspec
import fsspec.implementations.zip
from fsspec.implementations.local import LocalFileSystem

from xcube.core.byoa.constants import TEMP_FILE_PREFIX
from xcube.util.assertions import assert_given
from xcube.util.assertions import assert_instance
from xcube.util.jsonschema import JsonArraySchema
from xcube.util.jsonschema import JsonObject
from xcube.util.jsonschema import JsonObjectSchema
from xcube.util.jsonschema import JsonStringSchema
from xcube.util.temp import new_temp_dir
from xcube.util.temp import new_temp_file


class _FileSetDetails:
    def __init__(self,
                 fs: fsspec.AbstractFileSystem,
                 root: str,
                 local_path: Union[None, str]):
        """
        Stores file system details about a FileSet instance.
        Internal helper class.
        :param fs: file system
        :param root: root in file system
        :param local_path: the local or network path.
            None, if this is not a local path.
        """
        self._fs = fs
        self._root = root
        self._local_path = local_path
        self._mapper: Optional[fsspec.FSMap] = None

    @classmethod
    def new(cls,
            path: str,
            storage_params: Dict[str, Any] = None) -> '_FileSetDetails':
        try:
            fs, root = fsspec.core.url_to_fs(path,
                                             **(storage_params or {}))
        except (ImportError, OSError) as e:
            raise ValueError(f'Illegal file set {path!r}') from e
        local_path = root \
            if 'file' in fs.protocol or isinstance(fs, LocalFileSystem) \
            else None
        return _FileSetDetails(fs, root, local_path)

    @property
    def fs(self) -> fsspec.AbstractFileSystem:
        return self._fs

    @property
    def root(self) -> str:
        return self._root

    @property
    def local_path(self) -> Optional[str]:
        return self._local_path

    @property
    def mapper(self) -> fsspec.FSMap:
        if self._mapper is None:
            fs, root = self.fs, self.root
            if self.local_path \
                    and zipfile.is_zipfile(self.local_path):
                fs = fsspec.implementations.zip.ZipFileSystem(self.local_path)
                root = ''
            self._mapper = fsspec.FSMap(root, fs)
        return self._mapper


class FileSet(JsonObject):
    """
    A set of files that can found at some abstract root *path*.
    The *path* may identify local or remote filesystems.
    A filesystem may require specific *parameters*, e.g.
    user credentials.

    This implementation is based on the ```fsspec```
    package and the documentation about the different
    possible file systems and their specific parameters
    can be found at https://filesystem-spec.readthedocs.io/.

    Examples:

        FileSet('eurodatacube/my_processor',
                includes=['*.py'])

        FileSet('eurodatacube/my_processor.zip',
                excludes=['*.zarr'])

        FileSet('s3://eurodatacube/my_processor.zip',
                storage_params=dict(key='...',
                                    secret='...')

        FileSet('github://dcs4cop:xcube@v0.8.1/',
                storage_params=dict(username='...',
                                    token='ghp_...')

    If *includes* wildcard patterns are given, only files
    whose paths match any of the patterns are included.
    If *excludes* wildcard patterns are given, any file
    matching one of the patterns is excluded.
    The *includes*, if any, are applied before the
    *excludes*, if any.

    :param path: Root path of the file set.
    :param sub_path: An optional sub-path relative to *path*.
        If given, this is where the actual file sets starts.
        For example, the only top-level directory entry in a
        ZIP archive. In archive files from GitHub this would
        be typically "<gh_repo>-<gh-release-name>".
        Note: wildcard characters are not yet allowed.
    :param includes: Wildcard patterns used to include a file.
    :param excludes: Wildcard patterns used to exclude a file.
    :param storage_params: File system specific storage parameters.
    """

    def __init__(self,
                 path: str,
                 sub_path: str = None,
                 includes: Collection[str] = None,
                 excludes: Collection[str] = None,
                 storage_params: Dict[str, Any] = None):
        assert_instance(path, str, 'path')
        assert_given(path, 'path')
        if sub_path is not None:
            assert_instance(sub_path, str, 'sub_path')
        self._path = path
        self._sub_path = sub_path
        self._storage_params = dict(storage_params) \
            if storage_params is not None else None
        self._includes = list(includes) if includes is not None else None
        self._excludes = list(excludes) if excludes is not None else None
        # computed members
        self._include_patterns = _translate_patterns(includes or [])
        self._exclude_patterns = _translate_patterns(excludes or [])
        # cached, computed members
        self._details: Optional[_FileSetDetails] = None

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        """Get the JSON-schema for FileSet objects."""
        return JsonObjectSchema(
            properties=dict(
                path=JsonStringSchema(min_length=1),
                sub_path=JsonStringSchema(min_length=1),
                storage_params=JsonObjectSchema(additional_properties=True),
                includes=JsonArraySchema(
                    items=JsonStringSchema(min_length=1)
                ),
                excludes=JsonArraySchema(
                    items=JsonStringSchema(min_length=1
                                           )),
            ),
            additional_properties=False,
            required=['path'],
            factory=cls,
        )

    @property
    def path(self) -> str:
        """Get the root path."""
        return self._path

    @property
    def sub_path(self) -> Optional[str]:
        """Get the sub-path relative to *path*."""
        return self._sub_path

    @property
    def storage_params(self) -> Optional[Dict[str, Any]]:
        """
        Get optional parameters for the file system
        :attribute:path is referring to.
        """
        return self._storage_params

    @property
    def includes(self) -> Optional[List[str]]:
        """Wildcard patterns used to include a file."""
        return self._includes

    @property
    def excludes(self) -> Optional[List[str]]:
        """Wildcard patterns used to exclude a file."""
        return self._excludes

    def keys(self) -> Iterator[str]:
        """
        Get keys in this file set.
        A key is a normalized path relative to the root *path*.
        The forward slash "/" is used as path separator.
        """
        for key in self._get_details().mapper.keys():
            # print('-->', key)
            if self._accepts_key(key):
                yield key

    def get_local_path(self) -> Optional[str]:
        """
        Get the local path, if this file set is local,
        otherwise return None.
        """
        return self._get_details().local_path

    def is_local(self) -> bool:
        """
        Test whether this file set refers to a local file or directory.
        The test is made on this path's protocol.
        """
        if self._details is not None:
            return self._details.local_path is not None
        protocol, _ = fsspec.core.split_protocol(self._path)
        return protocol is None or protocol == 'file'

    def to_local(self) -> 'FileSet':
        """
        Turn this file set into a locally existing file set.
        """
        if self.is_local():
            return self

        details = self._get_details()
        fs, root = details.fs, details.root

        url_path = fsspec.core.strip_protocol(self.path)
        temp_file_suffix = ''
        for temp_file_suffix in reversed(url_path.split('/')):
            if temp_file_suffix != '':
                break

        if fs.isdir(root):
            temp_dir = new_temp_dir(prefix=TEMP_FILE_PREFIX,
                                    suffix=temp_file_suffix)

            # Note, actually want to use
            #    fs.get(root + "/*", temp_dir, recursive=True)
            # here, but fs.get() behaves unpredictably with
            # fsspec=2023.3.0 and s3fs=2023.3.0.
            # Sometimes it will create a new subdirectory in temp_dir
            # named after last path element of root, sometimes not.
            # Behaviour randomly changes if
            #   - temp_dir exists or not
            #   - root or temp_dir paths end with a slash
            #   - whether root or temp_dir are absolute path or not.
            #
            # See https://github.com/dcs4cop/xcube/issues/828
            #
            # Workaround used here is to manually download the directory.

            def get_files(source: str, target: str):
                source_items = fs.listdir(source, detail=True)
                for source_item in source_items:
                    source_path = source_item["name"]
                    source_type = source_item["type"]
                    source_name = source_path.replace("\\", "/") \
                        .split("/")[-1]
                    target_path = os.path.join(target, source_name)
                    if source_type == "file":
                        fs.get_file(source_path, target_path)
                    elif source_type == "directory":
                        os.mkdir(target_path)
                        get_files(source_path, target_path)

            get_files(root, temp_dir)

            return FileSet(temp_dir,
                           sub_path=self.sub_path,
                           includes=self.includes,
                           excludes=self.excludes)
        else:
            _, temp_file = new_temp_file(prefix=TEMP_FILE_PREFIX,
                                         suffix=temp_file_suffix)
            fs.get_file(root, temp_file)
            return FileSet(temp_file,
                           sub_path=self.sub_path,
                           includes=self.includes,
                           excludes=self.excludes)

    def is_local_dir(self) -> bool:
        """
        Test whether this file set refers to an existing local directory.
        """
        local_path = self.get_local_path()
        return os.path.isdir(local_path) \
            if local_path is not None else False

    def to_local_dir(self, dir_path: str = None) -> 'FileSet':
        """
        Convert this file set into a file set that refers to a
        directory in the local file system.

        The *dir_path* parameter is used only if this fileset
        does not already refer to an existing local directory.

        :param dir_path: An optional directory path.
            If not given, a temporary directory is created.
        :return: The file set representing the local directory.
        """
        file_set = self.to_local()
        if file_set.is_local_dir() \
                and dir_path is None \
                and not file_set.includes \
                and not file_set.excludes:
            return file_set
        return file_set._write_local_dir(dir_path)

    def is_local_zip(self):
        """
        Test whether this file set refers to an existing local ZIP archive.
        """
        local_path = self.get_local_path()
        return zipfile.is_zipfile(local_path) \
            if local_path is not None else False

    def to_local_zip(self, zip_path: str = None) -> 'FileSet':
        """
        Zip this file set and return it as a new local directory file set.
        If this is already a ZIP archive, return this file set.

        :param zip_path: An optional path for the new ZIP archive.
            If not given, a temporary file will be created.
        :return: The file set representing the local ZIP archive.
        """
        file_set = self.to_local()
        if file_set.is_local_zip() \
                and zip_path is None \
                and not file_set._includes \
                and not file_set._excludes:
            return file_set
        return file_set._write_local_zip(zip_path)

    def _write_local_dir(self, dir_path: Optional[str]) -> 'FileSet':
        """Write file set into local directory."""
        if dir_path:
            if not os.path.exists(dir_path):
                os.makedirs(dir_path, exist_ok=True)
            else:
                # We should assert an empty existing directory
                pass
        else:
            dir_path = new_temp_dir(prefix=TEMP_FILE_PREFIX)

        mapper = self._get_details().mapper

        sub_path = _normalize_sub_path(self.sub_path)

        for key in self.keys():
            file_path = os.path.join(dir_path,
                                     _key_to_path(key, sub_path))
            file_dir_path = os.path.dirname(file_path)
            if not os.path.exists(file_dir_path):
                os.makedirs(file_dir_path, exist_ok=True)
            with open(file_path, 'wb') as stream:
                stream.write(mapper[key])

        return FileSet(dir_path)

    def _write_local_zip(self, zip_path: Optional[str]) -> 'FileSet':
        """Write file set into local ZIP archive."""
        local_dir_path = self.get_local_path()

        if not zip_path:
            _, zip_path = new_temp_file(prefix=TEMP_FILE_PREFIX,
                                        suffix='.zip')

        sub_path = _normalize_sub_path(self.sub_path)

        with zipfile.ZipFile(zip_path, 'w') as zip_file:
            for key in self.keys():
                file_path = os.path.join(local_dir_path, key)
                zip_file.write(
                    file_path,
                    arcname=_strip_sub_path_from_key(key, sub_path)
                )

        return FileSet(zip_path, sub_path=self.sub_path)

    def _accepts_key(self, key: str) -> bool:
        if key == '':
            return False
        key = '/' + key.replace(os.path.sep, "/")
        return _key_matches(key, self._include_patterns, True) \
               and not _key_matches(key, self._exclude_patterns, False)

    def _get_details(self) -> _FileSetDetails:
        if self._details is None:
            self._details = _FileSetDetails.new(self.path,
                                                self.storage_params)
        return self._details

    def _get_fs(self) -> fsspec.AbstractFileSystem:
        return self._get_details().fs


def _key_matches(key: str,
                 patterns: List[re.Pattern],
                 empty_value: bool) -> bool:
    if not patterns:
        return empty_value
    for pattern in patterns:
        if pattern.match(key):
            return True
    return False


def _translate_patterns(patterns: Optional[Collection[str]]) \
        -> List[re.Pattern]:
    return [re.compile(fnmatch.translate(np))
            for p in patterns
            for np in _normalize_pattern(p)]


def _normalize_sub_path(sub_path: str) -> str:
    if sub_path:
        while sub_path.startswith('/'):
            sub_path = sub_path[1:]
        while sub_path.endswith('/'):
            sub_path = sub_path[:-1]
        if sub_path:
            sub_path += '/'
    return sub_path or ''


def _normalize_pattern(pattern: str) -> List[str]:
    pattern = pattern.replace(os.path.sep, '/')
    start_sep = pattern.startswith('/')
    end_sep = pattern.endswith('/')
    if start_sep and end_sep:
        return [pattern,
                f'{pattern}*']
    elif start_sep:
        return [pattern,
                f'{pattern}/*']
    elif end_sep:
        return [pattern,
                f'{pattern}*',
                f'*/{pattern}',
                f'*/{pattern}*']
    else:
        return [pattern,
                f'*/{pattern}',
                f'*/{pattern}/*']


def _strip_sub_path_from_key(key: str, sub_path: str):
    if key.startswith(sub_path):
        return key[len(sub_path):]
    return key


def _key_to_path(key: str, sub_path: str):
    key_path = key
    if sub_path:
        key_path = _strip_sub_path_from_key(key, sub_path)
    return key_path.replace('/', os.path.sep)



# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import datetime
import numbers
import traceback
import warnings
from typing import Optional, Sequence, Any, Tuple

import pandas as pd
import pyproj

from xcube.core.store import DataStorePool
from xcube.core.store import DatasetDescriptor
from xcube.core.store import VariableDescriptor
from xcube.util.assertions import assert_instance
from xcube.util.assertions import assert_true
from .describer import DatasetsDescriber
from ..config import CubeConfig
from ..error import CubeGeneratorError
from ..request import CubeGeneratorRequest
from ..response import CubeInfo


class CubeInformant:
    def __init__(self,
                 request: CubeGeneratorRequest,
                 store_pool: DataStorePool = None):
        assert_instance(request, CubeGeneratorRequest, name='request')
        if store_pool is not None:
            assert_instance(store_pool, DataStorePool, name='store_pool')
        self._request: CubeGeneratorRequest = request
        self._store_pool: Optional[DataStorePool] = store_pool
        self._dataset_descriptors: Optional[Sequence[DatasetDescriptor]] = None

    @property
    def input_dataset_descriptors(self) -> Sequence[DatasetDescriptor]:
        """Get dataset descriptors for inputs, lazily."""
        if self._dataset_descriptors is None:
            dataset_describer = DatasetsDescriber(
                self._request.input_configs,
                store_pool=self._store_pool
            )
            dataset_descriptors = dataset_describer.describe_datasets()
            if len(dataset_descriptors) > 1:
                warnings.warn('Only the first input will be recognised.')
            self._dataset_descriptors = dataset_descriptors
        return self._dataset_descriptors

    @property
    def first_input_dataset_descriptor(self) -> DatasetDescriptor:
        """Get dataset descriptor of first input, lazily."""
        return self.input_dataset_descriptors[0]

    @property
    def effective_cube_config(self) -> CubeConfig:
        cube_config, _, _ = self._compute_effective_cube_config()
        return cube_config

    def generate(self) -> CubeInfo:
        try:
            cube_config, resolved_crs, resolved_time_range = \
                 self._compute_effective_cube_config()
        except (TypeError, ValueError) as e:
            raise CubeGeneratorError(f'{e}', status_code=400) from e

        x_min, y_min, x_max, y_max = cube_config.bbox
        spatial_res = cube_config.spatial_res

        width = round((x_max - x_min) / spatial_res)
        height = round((y_max - y_min) / spatial_res)
        width = 2 if width < 2 else width
        height = 2 if height < 2 else height

        num_tiles_x = 1
        num_tiles_y = 1
        tile_width = width
        tile_height = height

        tile_size = cube_config.tile_size
        if tile_size is None and cube_config.chunks is not None:
            # TODO: this is just an assumption, with new
            #   Resampling module, use GridMapping
            #   to identify the actual names for the
            #   spatial tile dimensions.
            tile_size_x = cube_config.chunks.get('lon',
                                                 cube_config.chunks.get('x'))
            tile_size_y = cube_config.chunks.get('lat',
                                                 cube_config.chunks.get('y'))
            if tile_size_x and tile_size_y:
                tile_size = tile_size_x, tile_size_y

        if tile_size is not None:
            tile_width, tile_height = tile_size

            # TODO: this must be made common store logic
            if width > 1.5 * tile_width:
                num_tiles_x = _idiv(width, tile_width)
                width = num_tiles_x * tile_width

            # TODO: this must be made common store logic
            if height > 1.5 * tile_height:
                num_tiles_y = _idiv(height, tile_height)
                height = num_tiles_y * tile_height

        variable_names = cube_config.variable_names

        num_times = len(resolved_time_range)
        num_variables = len(variable_names)
        num_requests = num_variables \
                       * num_times \
                       * num_tiles_x * num_tiles_y
        # TODO: get original data types from dataset descriptors
        num_bytes_per_pixel = 4
        num_bytes = num_variables \
                    * num_times \
                    * (height * width * num_bytes_per_pixel)

        x_name, y_name = ('lon', 'lat') \
            if resolved_crs.is_geographic else ('x', 'y')

        data_id = self._request.output_config.data_id or 'unnamed'
        # TODO: get original variable descriptors from input dataset descriptors
        data_vars = {name: VariableDescriptor(name,
                                              dtype='float32',
                                              dims=('time', y_name, x_name))
                     for name in variable_names}
        dims = {'time': num_times, y_name: height, x_name: width}
        dataset_descriptor = DatasetDescriptor(
            data_id,
            crs=cube_config.crs,
            bbox=cube_config.bbox,
            spatial_res=cube_config.spatial_res,
            time_range=cube_config.time_range,
            time_period=cube_config.time_period,
            dims=dims,
            data_vars=data_vars
        )
        size_estimation = dict(
            image_size=[width, height],
            tile_size=[tile_width, tile_height],
            num_variables=num_variables,
            num_tiles=[num_tiles_x, num_tiles_y],
            num_requests=num_requests,
            num_bytes=num_bytes
        )

        return CubeInfo(dataset_descriptor=dataset_descriptor,
                        size_estimation=size_estimation)

    def _compute_effective_cube_config(self) \
            -> Tuple[CubeConfig, pyproj.crs.CRS, Sequence[pd.Timestamp]]:
        """
        Compute the effective cube configuration.

        This method reflects the behaviour of the LocalCubeGenerator
        that would normalize, tailor, and optionally resample a dataset
        based on the dataset descriptor and the cube configuration parameters,
        in the case that a store is not able to do so.

        :return: effective cube configuration.
        """

        request = self._request

        cube_config = request.cube_config \
            if request.cube_config is not None else CubeConfig()

        crs = cube_config.crs
        if crs is None:
            crs = self.first_input_dataset_descriptor.crs
            if crs is None:
                crs = 'WGS84'
        try:
            resolved_crs = pyproj.crs.CRS.from_string(crs)
        except (ValueError, pyproj.exceptions.CRSError) as e:
            raise ValueError(f'crs is invalid: {e}') from e

        bbox = cube_config.bbox
        if bbox is None:
            bbox = self.first_input_dataset_descriptor.bbox
        try:
            x1, y1, x2, y2 = bbox
        except (TypeError, ValueError):
            raise ValueError('bbox must be a tuple (x1, y1, x2, y2)')
        assert_instance(x1, numbers.Number, 'x1 of bbox')
        assert_instance(y1, numbers.Number, 'y1 of bbox')
        assert_instance(x2, numbers.Number, 'x2 of bbox')
        assert_instance(y2, numbers.Number, 'y2 of bbox')
        if resolved_crs.is_geographic and x2 < x1:
            x2 += 360
        bbox = x1, y1, x2, y2

        spatial_res = cube_config.spatial_res
        if spatial_res is None:
            spatial_res = self.first_input_dataset_descriptor.spatial_res
        assert_instance(spatial_res, numbers.Number, 'spatial_res')
        assert_true(spatial_res > 0, 'spatial_res must be positive')

        tile_size = cube_config.tile_size
        if tile_size is not None:
            try:
                tile_width, tile_height = tile_size
            except (TypeError, ValueError):
                raise ValueError('tile_size must be a'
                                 ' tuple (tile_width, tile_height)')
            tile_size = tile_width, tile_height

        time_range = cube_config.time_range
        if time_range is None:
            time_range = None, None
        start_ts, end_ts = _parse_time_range(time_range)
        if start_ts is None or end_ts is None:
            default_time_range = \
                self.first_input_dataset_descriptor.time_range
            if default_time_range is None:
                default_time_range = None, None
            default_start_ts, default_end_ts = \
                _parse_time_range(default_time_range)
            if start_ts is None:
                start_ts = default_start_ts
                if start_ts is None:
                    start_ts = pd.Timestamp(datetime.date.today())
            if end_ts is None:
                end_ts = default_end_ts
                if end_ts is None:
                    end_ts = pd.Timestamp(datetime.date.today()) \
                             + pd.Timedelta('1D')
        time_range = (start_ts.strftime("%Y-%m-%d"),
                      end_ts.strftime("%Y-%m-%d"))

        time_period = cube_config.time_period
        if time_period is None:
            time_period = self.first_input_dataset_descriptor.time_period
            if time_period is None:
                time_period = '1D'
        assert_instance(time_period, str, 'time_period')

        try:
            resolved_time_range = pd.date_range(start=start_ts,
                                                end=end_ts,
                                                freq=time_period)
        except ValueError as e:
            raise ValueError(f'invalid time_range or time_period: {e}') from e

        variable_names = cube_config.variable_names
        if variable_names is None:
            variable_names = list(
                self.first_input_dataset_descriptor.data_vars.keys()
            )

        return CubeConfig(
            variable_names=variable_names,
            crs=crs,
            bbox=bbox,
            spatial_res=spatial_res,
            tile_size=tile_size,
            time_range=time_range,
            time_period=time_period
        ), resolved_crs, resolved_time_range


def _idiv(x: int, y: int) -> int:
    return (x + y - 1) // y


def _parse_time_range(time_range: Any) \
        -> Tuple[Optional[pd.Timestamp], Optional[pd.Timestamp]]:
    try:
        start_date, end_date = time_range
    except (TypeError, ValueError):
        raise ValueError('time_range must be a tuple (start_date, end_date)')
    try:
        if start_date is not None:
            start_date = pd.Timestamp(start_date)
        if end_date is not None:
            end_date = pd.Timestamp(end_date)
    except ValueError as e:
        raise ValueError(f'time_range is invalid: {e}') from e
    return start_date, end_date



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from typing import List, Optional, Any, Mapping

from xcube.constants import LOG
from xcube.server.api import ApiError
from xcube.util.jsonschema import JsonArraySchema
from xcube.util.jsonschema import JsonBooleanSchema
from xcube.util.jsonschema import JsonObjectSchema
from xcube.util.jsonschema import JsonStringSchema

AUTHENTICATION_SCHEMA = JsonObjectSchema(
    properties=dict(
        Authority=JsonStringSchema(format='uri'),
        Domain=JsonStringSchema(description='Deprecated, use Authority'),
        Audience=JsonStringSchema(),
        Algorithms=JsonArraySchema(
            items=JsonStringSchema(default=["RS256"]),
        ),
        IsRequired=JsonBooleanSchema(default=False),
    ),
    required=[
        # Either "Authority" or "Domain" are required
        # "Authority",
        "Audience"
    ],
    additional_properties=False
)


class AuthConfig:
    @classmethod
    def get_schema(cls):
        return JsonObjectSchema(
            properties=dict(
                Authentication=AUTHENTICATION_SCHEMA
            ),
            factory=AuthConfig.from_dict
        )

    def __init__(self,
                 authority: str,
                 audience: str,
                 algorithms: List[str],
                 is_required: bool = False):
        self._authority = authority
        self._audience = audience
        self._algorithms = algorithms
        self._is_required = is_required

    @property
    def authority(self) -> str:
        return self._authority

    @property
    def _norm_authority(self) -> str:
        return self.authority[:-1] \
            if self.authority.endswith('/') \
            else self.authority

    @property
    def well_known_oid_config(self) -> str:
        return f"{self._norm_authority}/.well-known/openid-configuration"

    @property
    def well_known_jwks(self) -> str:
        return f"{self._norm_authority}/.well-known/jwks.json"

    @property
    def audience(self) -> str:
        return self._audience

    @property
    def algorithms(self) -> List[str]:
        return self._algorithms

    @property
    def is_required(self) -> bool:
        return self._is_required

    @classmethod
    def from_config(cls, config: Mapping[str, Any]) -> Optional['AuthConfig']:
        authentication = config.get('Authentication')
        if authentication is None:
            return None
        return cls.from_dict(authentication)

    @classmethod
    def from_dict(cls, config: Mapping[str, Any]) -> Optional['AuthConfig']:
        authority = config.get('Authority')
        domain = config.get('Domain')
        if domain:
            LOG.warning('Configuration parameter "Domain"'
                        ' in section "Authentication"'
                        ' has been deprecated. Use "Authority" instead.')
            if not authority:
                authority = f"https://{domain}"
        if not authority:
            raise ApiError.InvalidServerConfig(
                'Missing key "Authority" in section "Authentication"'
            )
        audience = config.get('Audience')
        if not audience:
            raise ApiError.InvalidServerConfig(
                'Missing key "Audience" in section "Authentication"'
            )
        algorithms = config.get('Algorithms', ['RS256'])
        if not algorithms:
            raise ApiError.InvalidServerConfig(
                'Value for key "Algorithms" in section'
                ' "Authentication" must not be empty'
            )
        is_required = config.get('IsRequired', False)
        return AuthConfig(authority,
                          audience,
                          algorithms,
                          is_required=is_required)



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube development team and contributors
# 
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import collections.abc
import warnings
from typing import Iterator, List

import zarr.storage

from xcube.util.assertions import assert_instance, assert_true


class CachedZarrStore(zarr.storage.Store):
    """A read-only Zarr store that is faster than
    *store* because it uses a writable *cache* store.

    The *cache* store is assumed to
    read values for a given key much faster than *store*.

    Note that iterating keys and containment checks are performed
    on *store* only.

    :param store: A Zarr store that is known
        to be slow in reading values.
    :param cache: A writable Zarr store that can
        read values faster than *store*.
    """

    _readable = True  # Because the base class is readable
    _listable = True  # Because the base class is listable
    _writeable = False  # Because this is not yet supported
    _erasable = False  # Because this is not yet supported

    def __init__(self,
                 store: collections.abc.MutableMapping,
                 cache: collections.abc.MutableMapping):
        assert_instance(store, collections.abc.MutableMapping, name="store")
        assert_instance(cache, collections.abc.MutableMapping, name="cache")
        if not isinstance(store, zarr.storage.BaseStore):
            store = zarr.storage.KVStore(store)
        if not isinstance(cache, zarr.storage.BaseStore):
            cache = zarr.storage.KVStore(cache)
        assert_true(store.is_readable(), message='store must be readable')
        assert_true(cache.is_readable(), message='cache must be readable')
        assert_true(cache.is_writeable(), message='cache must be writable')
        self._store = store
        self._cache = cache
        self._implement_op('listdir')
        self._implement_op('getsize')

    @property
    def store(self) -> zarr.storage.BaseStore:
        return self._store

    @property
    def cache(self) -> zarr.storage.BaseStore:
        return self._cache

    def _implement_op(self, op: str):
        if hasattr(self._store, op):
            assert hasattr(self, "_" + op)
            setattr(self, op, getattr(self, "_" + op))

    def _listdir(self, path: str = "") -> List[str]:
        # noinspection PyUnresolvedReferences
        return self._store.listdir(path=path)

    def _getsize(self, path: str) -> None:
        # noinspection PyBroadException
        try:
            # noinspection PyUnresolvedReferences
            size = self._cache.getsize(path)
        except BaseException:
            size = -1
        if size < 0:
            # noinspection PyUnresolvedReferences
            size = self._store.getsize(path)
        return size

    def __len__(self) -> int:
        return len(self._store)

    def __iter__(self) -> Iterator[str]:
        return iter(self._store)

    def __contains__(self, key: str):
        return key in self._store

    def __getitem__(self, key: str) -> bytes:
        try:
            return self._cache[key]
        except KeyError:
            pass
        value = self._store[key]
        # noinspection PyBroadException
        try:
            self._cache[key] = value
        except BaseException as e:
            warnings.warn(f"cache write failed for key {key!r}: {e}")
        return value

    def __setitem__(self, key: str, value: bytes) -> None:
        raise NotImplementedError()

    def __delitem__(self, key: str) -> None:
        raise NotImplementedError()



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.


import base64
import io
import os
from abc import ABC, abstractmethod
from functools import cached_property
from typing import Dict, Tuple, List, Optional

import fsspec
import matplotlib
import matplotlib.colors
import numpy as np
from PIL import Image
from deprecated import deprecated

from xcube.constants import LOG
from xcube.util.assertions import assert_instance

try:
    # noinspection PyPackageRequirements
    import cmocean.cm as cmo
except ImportError:
    cmo = None

OCEAN_PREFIX = "cmo"
ALPHA_SUFFIX = "_alpha"
REVERSED_SUFFIX = "_r"  # Used by ocm / Ocean

DEFAULT_CMAP_NAME = "viridis"


# Have colormaps separated into categories taken from
# https://matplotlib.org/stable/gallery/color/colormap_reference.html
# colormaps for ocean taken from
# https://matplotlib.org/cmocean/


class ColormapCategory:
    def __init__(self,
                 name: str,
                 desc: str,
                 cm_names: Optional[List[str]] = None):
        self.name = name
        self.desc = desc
        self.cm_names = cm_names


PERCEPTUALLY_UNIFORM_SEQUENTIAL_CATEGORY = ColormapCategory(
    'Perceptually Uniform Sequential',
    'For many applications, a perceptually uniform colormap'
    ' is the best choice - one in which equal steps in data are'
    ' perceived as equal steps in the color space',
    [
        'viridis', 'plasma', 'inferno', 'magma', 'cividis'
    ]
)

SEQUENTIAL_CATEGORY = ColormapCategory(
    'Sequential',
    'These colormaps are approximately monochromatic colormaps varying'
    ' smoothly between two color tones - usually from low saturation'
    ' (e.g. white) to high saturation (e.g. a bright blue).'
    ' Sequential colormaps are ideal for representing most scientific'
    ' data since they show a clear progression from'
    ' low-to-high values.',
    [
        'Greys', 'Purples', 'Blues', 'Greens', 'Oranges', 'Reds',
        'YlOrBr', 'YlOrRd', 'OrRd', 'PuRd', 'RdPu', 'BuPu',
        'GnBu', 'PuBu', 'YlGnBu', 'PuBuGn', 'BuGn', 'YlGn'
    ]
)

SEQUENTIAL_2_CATEGORY = ColormapCategory(
    'Sequential (2)',
    'Many of the values from the Sequential 2 plots are monotonically'
    ' increasing.',
    [
        'binary', 'gist_yarg', 'gist_gray', 'gray', 'bone', 'pink',
        'spring', 'summer', 'autumn', 'winter', 'cool', 'Wistia',
        'hot', 'afmhot', 'gist_heat', 'copper'
    ]
)

DIVERGING_CATEGORY = ColormapCategory(
    'Diverging',
    'These colormaps have a median value (usually light in color) and vary'
    ' smoothly to two different color tones at high and low values.'
    ' Diverging colormaps are ideal when your data has a median value'
    ' that is significant (e.g.  0, such that positive and negative'
    ' values are represented by different colors of the colormap).',
    [
        'PiYG', 'PRGn', 'BrBG', 'PuOr', 'RdGy', 'RdBu',
        'RdYlBu', 'RdYlGn', 'Spectral', 'coolwarm', 'bwr', 'seismic'
    ]
)

QUALITATIVE_CATEGORY = ColormapCategory(
    'Qualitative',
    'These colormaps vary rapidly in color.'
    ' Qualitative colormaps are useful for choosing'
    ' a set of discrete colors.',
    [
        'Pastel1', 'Pastel2', 'Paired', 'Accent',
        'Dark2', 'Set1', 'Set2', 'Set3',
        'tab10', 'tab20', 'tab20b', 'tab20c'
    ]
)

CYCLIC_CATEGORY = ColormapCategory(
    'Cyclic',
    'These colormaps are cyclic.',
    [
        'twilight', 'twilight_shifted', 'hsv'
    ]
)

MISCELLANEOUS_CATEGORY = ColormapCategory(
    'Miscellaneous',
    'Colormaps that don\'t fit into the categories above'
    ' or haven\'t been categorized yet.',
    [
        'flag', 'prism', 'ocean', 'gist_earth', 'terrain', 'gist_stern',
        'gnuplot', 'gnuplot2', 'CMRmap', 'cubehelix', 'brg',
        'gist_rainbow', 'rainbow', 'jet', 'turbo', 'nipy_spectral',
        'gist_ncar'
    ]
)

OCEAN_CATEGORY = ColormapCategory(
    'Ocean',
    'Colormaps for commonly-used oceanographic variables'
    ' (from module cmocean.cm).'
)

CUSTOM_CATEGORY = ColormapCategory(
    'Custom',
    'Custom colormaps, e.g. loaded from SNAP *.cpd files.',
)

TEMPLATE_CATEGORIES: Dict[str, ColormapCategory] = {
    c.name: c
    for c in (
        PERCEPTUALLY_UNIFORM_SEQUENTIAL_CATEGORY,
        SEQUENTIAL_CATEGORY,
        SEQUENTIAL_2_CATEGORY,
        DIVERGING_CATEGORY,
        QUALITATIVE_CATEGORY,
        CYCLIC_CATEGORY,
        OCEAN_CATEGORY,
        CUSTOM_CATEGORY,
        MISCELLANEOUS_CATEGORY,
    )
}

_CM_NAME_TO_TEMPLATE_CATEGORY = {
    cm_name: template_cat
    for template_cat in TEMPLATE_CATEGORIES.values()
    if template_cat.cm_names
    for cm_name in template_cat.cm_names
}


class Colormap:
    def __init__(self,
                 cm_name: str,
                 cat_name: Optional[str] = None,
                 cmap: Optional[matplotlib.colors.Colormap] = None,
                 cmap_reversed: Optional[matplotlib.colors.Colormap] = None,
                 cmap_alpha: Optional[matplotlib.colors.Colormap] = None,
                 norm: Optional[matplotlib.colors.Normalize] = None):
        self._cm_name = cm_name
        self._cat_name = cat_name
        self._cmap = cmap
        self._cmap_reversed = cmap_reversed
        self._cmap_alpha = cmap_alpha
        self._cmap_reversed_alpha = None
        self._cmap_png_base64: Optional[str] = None
        self._norm = norm

    @property
    def cm_name(self) -> str:
        return self._cm_name

    @property
    def cat_name(self) -> Optional[str]:
        return self._cat_name

    @cached_property
    def cmap(self) -> matplotlib.colors.Colormap:
        if self._cmap is None:
            self._cmap = matplotlib.colormaps[self.cm_name]
        return self._cmap

    @cached_property
    def cmap_alpha(self) -> matplotlib.colors.Colormap:
        if self._cmap_alpha is None:
            _, self._cmap_alpha = get_alpha_cmap(
                self.cm_name, self.cmap
            )
        return self._cmap_alpha

    @cached_property
    def cmap_reversed(self) -> matplotlib.colors.Colormap:
        if self._cmap_reversed is None:
            _, self._cmap_reversed = get_reverse_cmap(
                self.cm_name, self.cmap
            )
        return self._cmap_reversed

    @cached_property
    def cmap_reversed_alpha(self) -> matplotlib.colors.Colormap:
        if self._cmap_reversed_alpha is None:
            cm_name, cmap_reversed = get_reverse_cmap(
                self.cm_name, self.cmap
            )
            _, self._cmap_reversed_alpha = get_alpha_cmap(
                cm_name, cmap_reversed
            )
        return self._cmap_reversed_alpha

    @cached_property
    def cmap_png_base64(self) -> str:
        if self._cmap_png_base64 is None:
            self._cmap_png_base64 = get_cmap_png_base64(self.cmap)
        return self._cmap_png_base64

    @property
    def norm(self) -> Optional[matplotlib.colors.Normalize]:
        return self._norm


class ColormapProvider(ABC):
    @abstractmethod
    def get_cmap(self,
                 cm_name: str,
                 num_colors: Optional[int] = None) \
            -> Tuple[str, matplotlib.colors.Colormap]:
        """Get a colormap for the given *cm_name*.

        If *cm_name* is not available, the method may choose another
        colormap and return its name.

        A colormap provider may support colormaps that are

        * reversed (*cm_name* with suffix `"_r"`),
        * have alpha blending (*cm_name* with suffix `"_alpha"`),
        * both (*cm_name* with suffix `"_r_alpha"`).

        :param cm_name: Colormap name.
        :param num_colors: Optional number of colors, that is,
            the resolution of the colormap gradient.
        :return: A tuple comprising the base name
            of *cm_name* after striping any suffixes, and
            the colormap as an instance of ``matplotlib.colors.Colormap``.
        """


class ColormapRegistry(ColormapProvider):
    def __init__(self, *colormaps: Colormap):
        self._categories: Dict[str, ColormapCategory] = {}
        self._colormaps: Dict[str, Colormap] = {}
        # Add standard Matplotlib colormaps
        self._register_mpl_cmaps()
        # Add Ocean colormaps, if any
        self._register_ocm_cmaps()
        # Add custom colormaps, if any
        for colormap in colormaps:
            self.register_colormap(colormap)

    @property
    def categories(self) -> Dict[str, ColormapCategory]:
        return self._categories

    @property
    def colormaps(self) -> Dict[str, Colormap]:
        return self._colormaps

    def register_colormap(self, colormap: Colormap):
        cat_name = colormap.cat_name or MISCELLANEOUS_CATEGORY.name
        template_cat = TEMPLATE_CATEGORIES.get(cat_name)
        if template_cat is None:
            raise ValueError(f"Unknown colormap category {cat_name!r}")
        category = self._get_category(template_cat)
        category.cm_names.append(colormap.cm_name)
        self._colormaps[colormap.cm_name] = colormap

    def find_colormap(self, cm_name: str) -> Optional[Colormap]:
        assert_instance(cm_name, str, name="cm_name")
        # Strips suffixes from cm_name
        cm_name, _, _ = parse_cm_name(cm_name)
        return self._colormaps.get(cm_name)

    def get_cmap(self,
                 cm_name: str,
                 num_colors: Optional[int] = None) \
            -> Tuple[str, matplotlib.colors.Colormap]:
        assert_instance(cm_name, str, name="cm_name")
        if num_colors is not None:
            assert_instance(num_colors, int, name="num_colors")

        cm_name, reverse, alpha = parse_cm_name(cm_name)

        colormap = self._colormaps.get(cm_name)
        if colormap is None:
            cm_name = DEFAULT_CMAP_NAME
            colormap = self._colormaps[cm_name]

        if reverse and alpha:
            cmap = colormap.cmap_reversed_alpha
        elif reverse:
            cmap = colormap.cmap_reversed
        elif alpha:
            cmap = colormap.cmap_alpha
        else:
            cmap = colormap.cmap
        if num_colors is not None:
            try:
                # noinspection PyProtectedMember
                cmap = cmap._resample(num_colors)
            except (ValueError, AttributeError, NotImplementedError):
                pass
        return cm_name, cmap

    def to_json(self) -> List:
        result = []
        # Loop through TEMPLATE_CATEGORIES to preserve category order
        for cat_name in TEMPLATE_CATEGORIES.keys():
            category = self._categories.get(cat_name)
            if category is not None:
                result.append([
                    category.name,
                    category.desc,
                    [
                        [cm_name, self._colormaps[cm_name].cmap_png_base64]
                        for cm_name in category.cm_names
                    ]
                ])
        return result

    def _get_category(self, template_cat: ColormapCategory):
        category = self._categories.get(template_cat.name)
        if category is None:
            category = ColormapCategory(template_cat.name,
                                        template_cat.desc,
                                        cm_names=[])
            self._categories[template_cat.name] = category
        return category

    def _register_mpl_cmaps(self):
        for cm_name, cmap in matplotlib.colormaps.items():
            if cm_name.startswith(OCEAN_PREFIX):
                # We have a separate category for OCEAN maps
                continue
            if cm_name.endswith(REVERSED_SUFFIX):
                # We register reversed maps, if any, separately
                continue
            template_cat = _CM_NAME_TO_TEMPLATE_CATEGORY.get(cm_name)
            if template_cat is None:
                template_cat = MISCELLANEOUS_CATEGORY
            self.register_colormap(Colormap(
                cm_name,
                cat_name=template_cat.name,
                cmap=cmap,
                cmap_reversed=matplotlib.colormaps.get(
                    cm_name + REVERSED_SUFFIX
                )
            ))

    def _register_ocm_cmaps(self):
        if cmo is None:
            return
        for cm_name in cmo.__dict__.keys():
            if not isinstance(getattr(cmo, cm_name),
                              matplotlib.colors.Colormap) \
                    or cm_name.endswith(REVERSED_SUFFIX):
                continue
            self.register_colormap(Colormap(
                cm_name,
                cat_name=OCEAN_CATEGORY.name,
                cmap=getattr(cmo, cm_name),
                cmap_reversed=getattr(cmo,
                                      cm_name + REVERSED_SUFFIX,
                                      None)
            ))


def parse_cm_name(cm_name) -> Tuple[str, bool, bool]:
    alpha = cm_name.endswith(ALPHA_SUFFIX)
    if alpha:
        cm_name = cm_name[0:-len(ALPHA_SUFFIX)]
    reverse = cm_name.endswith(REVERSED_SUFFIX)
    if reverse:
        cm_name = cm_name[0:-len(REVERSED_SUFFIX)]
    return cm_name, reverse, alpha


def get_reverse_cmap(cm_name: str, cmap: matplotlib.colors.Colormap) \
        -> Tuple[str, matplotlib.colors.Colormap]:
    new_cm_name = cm_name + REVERSED_SUFFIX
    return new_cm_name, cmap.reversed(name=new_cm_name)


def get_alpha_cmap(cm_name: str, cmap: matplotlib.colors.Colormap) \
        -> Tuple[str, matplotlib.colors.Colormap]:
    """Add extra colormaps with alpha gradient.
    See http://matplotlib.org/api/colors_api.html
    """
    new_cm_name = cm_name + ALPHA_SUFFIX

    # noinspection SpellCheckingInspection
    if isinstance(cmap, matplotlib.colors.LinearSegmentedColormap) \
            and hasattr(cmap, "_segmentdata"):
        # noinspection PyProtectedMember
        new_segment_data = dict(cmap._segmentdata)
        # let alpha increase from 0.0 to 0.5
        new_segment_data['alpha'] = ((0.0, 0.0, 0.0),
                                     (0.5, 1.0, 1.0),
                                     (1.0, 1.0, 1.0))
        return new_cm_name, matplotlib.colors.LinearSegmentedColormap(
            new_cm_name, new_segment_data
        )

    if isinstance(cmap, matplotlib.colors.ListedColormap):
        new_colors = list(cmap.colors)
        a_slope = 2.0 / cmap.N
        a = 0
        for i in range(len(new_colors)):
            # noinspection PyUnresolvedReferences
            new_color = new_colors[i]
            if not isinstance(new_color, str):
                if len(new_color) == 3:
                    r, g, b = new_color
                    new_colors[i] = r, g, b, a
                elif len(new_color) == 4:
                    r, g, b, a_old = new_color
                    new_colors[i] = r, g, b, min(a, a_old)
            a += a_slope
            if a > 1.0:
                a = 1.0
        return new_cm_name, matplotlib.colors.ListedColormap(
            new_colors, name=new_cm_name
        )

    LOG.warning(
        f'Could not create alpha colormap for cmap of type {type(cmap)}'
    )

    return cm_name, cmap


def get_cmap_png_base64(cmap: matplotlib.colors.Colormap) -> str:
    gradient = np.linspace(0, 1, 256)
    gradient = np.vstack((gradient,))
    image_data = cmap(gradient, bytes=True)
    image = Image.fromarray(image_data, 'RGBA')

    # For testing, do:
    # stream = io.FileIO('../cmaps/' + cmap_name + '.png', 'wb')
    # image.save(stream, format='PNG')
    # stream.close()

    stream = io.BytesIO()
    image.save(stream, format='PNG')
    cbar_png_bytes = stream.getvalue()
    stream.close()

    cbar_png_data = base64.b64encode(cbar_png_bytes)
    cbar_png_bytes = cbar_png_data.decode('unicode_escape')

    return cbar_png_bytes


def load_custom_colormap(custom_colormap_path: str) -> Optional[Colormap]:
    try:
        # Currently, we only support SNAP *.cpd files
        colormap = load_snap_cpd_colormap(custom_colormap_path)
        LOG.info(f'Loaded custom colormap'
                 f' {custom_colormap_path!r}')
        return colormap
    except FileNotFoundError:
        LOG.error(f"Missing custom colormap file:"
                  f" {custom_colormap_path}")
    except ValueError as e:
        LOG.error(f'Detected invalid custom colormap'
                  f' {custom_colormap_path!r}: {e}',
                  exc_info=True)
    return None


def load_snap_cpd_colormap(snap_cpd_path: str) -> Colormap:
    points, log_scaled = _parse_snap_cpd_file(snap_cpd_path)
    samples = [sample for sample, _ in points]
    vmin, vmax = min(samples), max(samples)
    if log_scaled:
        norm = matplotlib.colors.LogNorm(vmin, vmax)
    else:
        norm = matplotlib.colors.Normalize(vmin, vmax)
    colors = list(zip(map(norm, samples),
                      ['#%02x%02x%02x' % color for _, color in points]))
    cm_name, _ = os.path.splitext(os.path.basename(snap_cpd_path))
    return Colormap(
        cm_name,
        cat_name=CUSTOM_CATEGORY.name,
        cmap=matplotlib.colors.LinearSegmentedColormap.from_list(
            cm_name, colors
        ),
        norm=norm
    )


Sample = float
Color = Tuple[int, int, int]
Palette = List[Tuple[Sample, Color]]
LogScaled = bool


def _parse_snap_cpd_file(cpd_file_path: str) \
        -> Tuple[Palette, LogScaled]:
    with fsspec.open(cpd_file_path, mode="r") as f:

        illegal_format_msg = f"Illegal SNAP *.cpd format: {cpd_file_path}"

        entries: Dict[str, str] = dict()
        for line in f.readlines():
            line = line.strip()
            if line and not line.startswith("#"):
                try:
                    k, v = line.split("=", maxsplit=1)
                except ValueError:
                    raise ValueError(illegal_format_msg)
                entries[k.strip()] = v.strip()

        for keyword in ("autoDistribute", "isLogScaled"):
            if keyword in entries:
                LOG.warning(f"Unrecognized keyword {keyword!r}"
                            f" in custom colormap {cpd_file_path}")

        # Uncomment, as soon as we know how to deal with this:
        # log_scaled = entries.get("isLogScaled") == "true"
        log_scaled = False

        try:
            num_points = int(entries.get("numPoints"))
        except ValueError:
            raise ValueError(illegal_format_msg)

        points = []
        for i in range(num_points):
            try:
                r, g, b = map(int, entries.get(f"color{i}", "").split(","))
                sample = float(entries.get(f"sample{i}"))
            except ValueError:
                raise ValueError(illegal_format_msg)
            points.append((sample, (r, g, b)))

        return points, log_scaled


############################################################################
# Deprecated API

_REGISTRY = None
_REGISTRY_JSON = None


@deprecated(reason="Colormaps in xcube are no longer managed globally. Use"
                   " xcube.util.cmaps.ColormapRegistry.to_json() instead.",
            version="0.13.0")
def get_cmaps() -> List:
    """
    Return a JSON-serializable tuple containing records of the form:
     (<cmap-category>, <cmap-category-description>, <cmap-tuples>),
    where <cmap-tuples> is a tuple containing records of the form
    (<cmap-name>, <cbar-png-bytes>), and where
    <cbar-png-bytes> are encoded PNG images of size 256 x 2 pixels,
    :return: all known matplotlib color maps
    """
    registry = _get_registry()
    global _REGISTRY_JSON
    if _REGISTRY_JSON is None:
        _REGISTRY_JSON = registry.to_json()
    return _REGISTRY_JSON


# noinspection PyUnusedLocal
@deprecated(reason="Colormaps in xcube are no longer managed globally. Use"
                   " xcube.util.cmaps.ColormapRegistry.get_cmap() instead.",
            version="0.13.0")
def get_cmap(cmap_name: str,
             default_cmap_name: str = "viridis",
             num_colors: Optional[int] = None) \
        -> Tuple[str, matplotlib.colors.Colormap]:
    """
    Get color mapping for color bar name *cmap_name*.

    If *num_colors* is a positive integer,
    a resampled mapping will be returned that contains *num_colors*
    color entries.

    If *cmap_name* is not defined, *default_cmap_name* is used.
    If *default_cmap_name* is undefined too, a ValueError is raised.

    Otherwise, a tuple (actual_cmap_name, cmap) is returned.

    :param cmap_name: Color bar name.
    :param num_colors: Number of colours in returned color mapping.
    :param default_cmap_name: Default color bar name. (Ignored)
    :return: A tuple (actual_cmap_name, cmap).
    """
    registry = _get_registry()
    return registry.get_cmap(cmap_name,
                             num_colors=num_colors)


@deprecated(reason="Colormaps in xcube are no longer managed globally."
                   " This function is obsolete.",
            version="0.13.0")
def ensure_cmaps_loaded():
    """
    Loads all color maps from matplotlib and registers additional ones,
    if not done before.
    """
    _get_registry()


def _get_registry() -> ColormapRegistry:
    global _REGISTRY
    if _REGISTRY is None:
        _REGISTRY = ColormapRegistry()
    return _REGISTRY



# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Type, Dict, Optional, Any, Sequence

import fsspec

from .accessor import FsAccessor
from .accessor import FsDataAccessor
from .impl.dataset import DatasetGeoTiffFsDataAccessor
from .impl.dataset import DatasetNetcdfFsDataAccessor
from .impl.dataset import DatasetZarrFsDataAccessor
from .impl.fs import AzureFsAccessor
from .impl.fs import FileFsAccessor
from .impl.fs import FtpFsAccessor
from .impl.fs import MemoryFsAccessor
from .impl.fs import S3FsAccessor
from .impl.geodataframe import GeoDataFrameGeoJsonFsDataAccessor
from .impl.geodataframe import GeoDataFrameShapefileFsDataAccessor
from .impl.geotiff import MultiLevelDatasetGeoTiffFsDataAccessor
from .impl.mldataset import DatasetLevelsFsDataAccessor
from .impl.mldataset import MultiLevelDatasetLevelsFsDataAccessor
from .store import FsDataStore
from ..assertions import assert_valid_params
from ..error import DataStoreError

############################################
# FsAccessor

_FS_ACCESSOR_CLASSES: Dict[str, Type[FsAccessor]] = {}


def register_fs_accessor_class(
        fs_accessor_class: Type[FsAccessor]
):
    """
    Register a concrete filesystem accessor class.

    :param fs_accessor_class: a concrete class
        that extends :class:FsAccessor.
    """
    protocol = fs_accessor_class.get_protocol()
    _FS_ACCESSOR_CLASSES[protocol] = fs_accessor_class


for cls in (
        AzureFsAccessor,
        FileFsAccessor,
        FtpFsAccessor,
        MemoryFsAccessor,
        S3FsAccessor,
):
    register_fs_accessor_class(cls)


def get_fs_accessor_class(protocol: str) -> Type[FsAccessor]:
    """
    Get the class for a filesystem accessor.

    :param protocol: The filesystem protocol,
        for example "file", "s3", "memory".
    :return: A class that derives from :class:FsAccessor
    """
    fs_accessor_class = _FS_ACCESSOR_CLASSES.get(protocol)
    if fs_accessor_class is None:
        try:
            fsspec.get_filesystem_class(protocol)
        except ImportError as e:
            raise DataStoreError(
                f'Filesystem for protocol {protocol!r}'
                f' is not installed or requires additional packages'
            ) from e
        except ValueError as e:
            raise DataStoreError(
                f'Filesystem not found for protocol {protocol!r}'
            ) from e

        class FsAccessorClass(FsAccessor):
            @classmethod
            def get_protocol(cls) -> str:
                return protocol

        fs_accessor_class = FsAccessorClass
    return fs_accessor_class


############################################
# FsDataAccessor

_FS_DATA_ACCESSOR_CLASSES: Dict[str, Type[FsDataAccessor]] = {}


def register_fs_data_accessor_class(
        fs_data_accessor_class: Type[FsDataAccessor]
):
    """
    Register an abstract filesystem data accessor class.

    Such data accessor classes are used to dynamically
    construct concrete data store classes by combining
    them with a concrete :class:FsAccessor.

    :param fs_data_accessor_class: an abstract class
        that extends :class:FsDataAccessor.
    """
    data_type = fs_data_accessor_class.get_data_type()
    format_id = fs_data_accessor_class.get_format_id()
    key = f'{data_type.alias}:{format_id}'
    _FS_DATA_ACCESSOR_CLASSES[key] = fs_data_accessor_class


for cls in (
        DatasetZarrFsDataAccessor,
        DatasetNetcdfFsDataAccessor,
        DatasetGeoTiffFsDataAccessor,
        DatasetLevelsFsDataAccessor,
        MultiLevelDatasetGeoTiffFsDataAccessor,
        MultiLevelDatasetLevelsFsDataAccessor,
        GeoDataFrameShapefileFsDataAccessor,
        GeoDataFrameGeoJsonFsDataAccessor,
):
    register_fs_data_accessor_class(cls)


def get_fs_data_accessor_class(protocol: str,
                               data_type_alias: str,
                               format_id: str) -> Type[FsDataAccessor]:
    """
    Get the class for a filesystem data accessor.

    :param protocol: The filesystem protocol,
        for example "file", "s3", "memory".
    :param data_type_alias: The data type alias name,
        for example "dataset", "geodataframe".
    :param format_id: The format identifier,
        for example "zarr", "geojson".
    :return: A class that derives from :class:FsAccessor
    """
    accessor_id = f'{data_type_alias}:{format_id}'
    data_accessor_class = _FS_DATA_ACCESSOR_CLASSES.get(accessor_id)
    if data_accessor_class is None:
        raise DataStoreError(f'Combination of data type {data_type_alias!r}'
                             f' and format {format_id!r} is not supported')

    fs_accessor_class = get_fs_accessor_class(protocol)

    class FsDataAccessorClass(fs_accessor_class, data_accessor_class):
        pass

    # Should we set __name_ and __doc__ properties here?

    return FsDataAccessorClass


############################################
# FsDataStore

def get_fs_data_store_class(protocol: str) -> Type[FsDataStore]:
    """
    Get the class for of a filesystem-based data store.

    :param protocol: The filesystem protocol,
        for example "file", "s3", "memory".
    :return: A class that derives from :class:FsDataStore
    """
    fs_accessor_class = get_fs_accessor_class(protocol)

    class FsDataStoreClass(fs_accessor_class, FsDataStore):
        pass

    # Should we set set __name_ and __doc__ properties here?

    return FsDataStoreClass


def new_fs_data_store(
        protocol: str,
        root: str = '',
        max_depth: Optional[int] = 1,
        read_only: bool = False,
        includes: Optional[Sequence[str]] = None,
        excludes: Optional[Sequence[str]] = None,
        storage_options: Dict[str, Any] = None
) -> FsDataStore:
    """
    Create a new instance of a filesystem-based data store.

    The data store is capable of filtering the data identifiers reported
    by ``get_data_ids()``. For this purpose the optional keywords
    `excludes` and `includes` are used which can both take the form of
    a wildcard pattern or a sequence of wildcard patterns:

    * ``excludes``: if given and if any pattern matches the identifier,
      the identifier is not reported.
    * ``includes``: if not given or if any pattern matches the identifier,
      the identifier is reported.

    :param protocol: The filesystem protocol,
        for example "file", "s3", "memory".
    :param root: Root or base directory.
        Defaults to "".
    :param max_depth: Maximum recursion depth. None means limitless.
        Defaults to 1.
    :param read_only: Whether this is a read-only store.
        Defaults to False.
    :param includes: Optional sequence of wildcards that include
        certain filesystem paths. Affects the data identifiers (paths)
        returned by `get_data_ids()`. By default, all paths are included.
    :param excludes: Optional sequence of wildcards that exclude
        certain filesystem paths. Affects the data identifiers (paths)
        returned by `get_data_ids()`. By default, no paths are excluded.
    :param storage_options: Options specific to the underlying filesystem
        identified by *protocol*.
        Used to instantiate the filesystem.
    :return: A new data store instance of type :class:FsDataStore.
    """
    fs_data_store_class = get_fs_data_store_class(protocol)
    store_params_schema = fs_data_store_class.get_data_store_params_schema()
    store_params = {k: v
                    for k, v in dict(root=root,
                                     max_depth=max_depth,
                                     read_only=read_only,
                                     includes=includes,
                                     excludes=excludes,
                                     storage_options=storage_options).items()
                    if v is not None}
    assert_valid_params(store_params,
                        name='store_params',
                        schema=store_params_schema)
    return fs_data_store_class(**store_params)



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube development team and contributors
# 
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import collections.abc
import threading
import warnings
from typing import Optional

import xarray as xr

from xcube.constants import LOG
from xcube.util.assertions import assert_instance


@xr.register_dataset_accessor('zarr_store')
class ZarrStoreHolder:
    """Represents a xarray dataset property ``zarr_store``.

    It is used to permanently associate a dataset with its
    Zarr store, which would otherwise not be possible.

    In xcube server, we use the new property to expose
    datasets via the S3 emulation API.

    For that concept to work, datasets must be associated
    with their Zarr stores explicitly.
    Therefore, the xcube data store framework sets the
    Zarr stores of datasets after opening them ``xr.open_zarr()``:

    ```python
    dataset = xr.open_zarr(zarr_store, **open_params)
    dataset.zarr_store.set(zarr_store)
    ```

    Note, that the dataset may change after the Zarr store has been set,
    so that the dataset and its Zarr store are no longer in sync.
    This may be an issue and limit the application of the new property.

    :param dataset: The xarray dataset that is
        associated with a Zarr store.
    """

    def __init__(self, dataset: xr.Dataset):
        self._dataset = dataset
        self._zarr_store: Optional[collections.abc.MutableMapping] = None
        self._lock = threading.RLock()

    def get(self) -> collections.abc.MutableMapping:
        """Get the Zarr store of a dataset.
        If no Zarr store has been set, the method will use
        ``GenericZarrStore.from_dataset()`` to create and set
        one.

        :return: The Zarr store.
        """
        if self._zarr_store is None:
            # Double-checked locking pattern
            with self._lock:
                if self._zarr_store is None:
                    from xcube.core.zarrstore import GenericZarrStore
                    self._zarr_store = GenericZarrStore.from_dataset(
                        self._dataset
                    )
                    source = self._dataset.encoding.get("source", "?")
                    LOG.warning(f"dataset {source!r} is assigned a"
                                f" GenericZarrStore which may introduce"
                                f" performance penalties")
        return self._zarr_store

    def set(self, zarr_store: collections.abc.MutableMapping) -> None:
        """Set the Zarr store of a dataset.
        :param zarr_store: The Zarr store.
        """
        assert_instance(zarr_store,
                        collections.abc.MutableMapping,
                        name='zarr_store')
        with self._lock:
            self._zarr_store = zarr_store

    def reset(self) -> None:
        """Resets the Zarr store."""
        with self._lock:
            self._zarr_store = None



# The MIT License (MIT)
# Copyright (c) 2019 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import functools
import math

import numpy as np
import xarray as xr

from xcube.core.maskset import MaskSet
from xcube.util.config import NameDictPairList
from xcube.util.config import to_resolved_name_dict_pairs
from xcube.util.expression import compute_array_expr


def evaluate_dataset(dataset: xr.Dataset,
                     processed_variables: NameDictPairList = None,
                     errors: str = 'raise') -> xr.Dataset:
    """
    Compute new variables or mask existing variables in *dataset*
    by the evaluation of Python expressions, that may refer to other
    existing or new variables.
    Returns a new dataset that contains the old and new variables,
    where both may bew now masked.

    Expressions may be given by attributes of existing variables in
    *dataset* or passed a via the *processed_variables* argument
    which is a sequence of variable name / attributes tuples.

    Two types of expression attributes are recognized in the attributes:

    1. The attribute ``expression`` generates
       a new variable computed from its attribute value.
    2. The attribute ``valid_pixel_expression`` masks out
       invalid variable values.

    In both cases the attribuite value must be a string that forms
    a valid Python expression that can reference any other preceding
    variables by name.
    The expression can also reference any flags defined by another
    variable according the their CF attributes ``flag_meaning``
    and ``flag_values``.

    Invalid variable values may be masked out using the value the
    ``valid_pixel_expression`` attribute whose value should form
    a Boolean Python expression. In case, the expression
    returns zero or false, the value of the ``_FillValue`` attribute
    or NaN will be used in the new variable.

    Other attributes will be stored as variable metadata as-is.

    :param dataset: A dataset.
    :param processed_variables: Optional list of variable
        name-attributes pairs that will processed in the given order.
    :param errors: How to deal with errors while evaluating expressions.
           May be be one of "raise", "warn", or "ignore".
    :return: new dataset with computed variables
    """

    if processed_variables:
        processed_variables = to_resolved_name_dict_pairs(
            processed_variables, dataset, keep=True
        )
    else:
        var_names = list(dataset.data_vars)
        var_names = sorted(var_names,
                           key=functools.partial(_get_var_sort_key, dataset))
        processed_variables = [(var_name, None) for var_name in var_names]

    # Initialize namespace with some constants and modules
    namespace = dict(NaN=np.nan, PI=math.pi, np=np, xr=xr)
    # Now add all mask sets and variables
    for var_name in dataset.data_vars:
        var = dataset[var_name]
        if MaskSet.is_flag_var(var):
            namespace[var_name] = MaskSet(var)
        else:
            namespace[var_name] = var

    for var_name, var_props in processed_variables:
        if var_name in dataset.data_vars:
            # Existing variable
            var = dataset[var_name]
            if var_props:
                var_props_temp = var_props
                var_props = dict(var.attrs)
                var_props.update(var_props_temp)
            else:
                var_props = dict(var.attrs)
        else:
            # Computed variable
            var = None
            if var_props is None:
                var_props = dict()

        do_load = var_props.get('load', False)

        expression = var_props.get('expression')
        if expression:
            # Compute new variable
            computed_array = compute_array_expr(expression,
                                                namespace=namespace,
                                                result_name=f'{var_name!r}',
                                                errors=errors)
            if computed_array is not None:
                if hasattr(computed_array, 'attrs'):
                    var = computed_array
                    var.attrs.update(var_props)
                if do_load:
                    computed_array.load()
                namespace[var_name] = computed_array

        valid_pixel_expression = var_props.get('valid_pixel_expression')
        if valid_pixel_expression:
            # Compute new mask for existing variable
            if var is None:
                raise ValueError(f'undefined variable {var_name!r}')
            valid_mask = compute_array_expr(
                valid_pixel_expression,
                namespace=namespace,
                result_name=f'valid mask for {var_name!r}',
                errors=errors
            )
            if valid_mask is not None:
                masked_var = var.where(valid_mask)
                if hasattr(masked_var, 'attrs'):
                    masked_var.attrs.update(var_props)
                if do_load:
                    masked_var.load()
                namespace[var_name] = masked_var

    computed_dataset = dataset.copy()
    for name, value in namespace.items():
        if isinstance(value, xr.DataArray):
            computed_dataset[name] = value

    return computed_dataset


def _get_var_sort_key(dataset: xr.Dataset, var_name: str):
    # noinspection SpellCheckingInspection
    attrs = dataset[var_name].attrs
    a1 = attrs.get('expression')
    a2 = attrs.get('valid_pixel_expression')
    v1 = 10 * len(a1) if a1 is not None else 0
    v2 = 100 * len(a2) if a2 is not None else 0
    return v1 + v2



# The MIT License (MIT)
# Copyright (c) 2023 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import pathlib
from typing import Type, Iterator

import fsspec
from fsspec.implementations.local import LocalFileSystem


def is_local_fs(fs: fsspec.AbstractFileSystem) -> bool:
    """
    Check whether *fs* is a local filesystem.
    """
    return 'file' in fs.protocol or isinstance(fs, LocalFileSystem)


def get_fs_path_class(fs: fsspec.AbstractFileSystem) \
        -> Type[pathlib.PurePath]:
    """
    Get the appropriate ``pathlib.PurePath`` class for the filesystem *fs*.
    """
    if is_local_fs(fs):
        # Will return PurePosixPath or a PureWindowsPath object
        return pathlib.PurePath
    # PurePosixPath for non-local filesystems such as S3,
    # so we force use of forward slashes as separators
    return pathlib.PurePosixPath


def resolve_path(path: pathlib.PurePath) -> pathlib.PurePath:
    """
    Resolve "." and ".." occurrences in *path* without I/O and
    return a new path.
    """
    reversed_parts = reversed(path.parts)
    reversed_norm_parts = list(_resolve_path_impl(reversed_parts))
    parts = reversed(reversed_norm_parts)
    return type(path)(*parts)


def _resolve_path_impl(reversed_parts: Iterator[str]):
    skips = False
    for part in reversed_parts:
        if part == '.':
            continue
        elif part == '..':
            skips += 1
            continue
        if skips == 0:
            yield part
        elif skips > 0:
            skips -= 1
    if skips != 0:
        raise ValueError('cannot resolve path, misplaced ".."')



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube development team and contributors
# 
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import abc
from typing import Optional, Dict, Union, Sequence, List

from xcube.util.assertions import assert_given, assert_false


class Node(abc.ABC):
    """An abstract XML node."""

    @abc.abstractmethod
    def add(self, *elements: 'Element') -> 'Node':
        """
        Adds child elements to this element.
        :param elements: Child elements
        """
        pass

    @abc.abstractmethod
    def to_xml(self, indent: int = 2) -> str:
        """
        Converts this node into valid XML text using UTF-8 encoding.
        :param indent: Indent in spaces. Defaults to 2.
        """
        pass

    def __str__(self):
        """Calls to_xml() with default arguments."""
        return self.to_xml()


class Element(Node):
    """
    An XML element node.

    :param tag: Tag name
    :param attrs: Attributes
    :param text: Text
    :param elements: Child elements
    """

    def __init__(self,
                 tag: str,
                 attrs: Optional[Dict[str, str]] = None,
                 text: Optional[Union[str, Sequence[str]]] = None,
                 elements: Optional[Sequence['Element']] = None):
        assert_given(tag, name='tag')
        assert_false(text and elements,
                     message='text and elements are mutually exclusive')
        self._tag = tag
        self._text = text
        self._attrs = dict(attrs) if attrs else {}
        self._elements = list(elements) if elements else []

    def add(self, *elements: 'Element') -> 'Element':
        self._elements.extend(elements)
        return self

    def to_xml(self, indent: int = 2) -> str:
        lines = []
        self._to_xml(indent, 0, lines)
        return '\n'.join(lines)

    def _to_xml(self, indent: int, level: int, lines: List[str]):
        tab = indent * ' '
        tabs = level * tab
        line = f'{tabs}<{self._tag}'
        attrs = self._attrs
        text = self._text
        elements = self._elements
        if attrs:
            for k, v in attrs.items():
                line += f' {k}="{v}"'
        if text:
            if isinstance(text, str):
                lines.append(f'{line}>{text}</{self._tag}>')
            else:
                lines.append(line + '>')
                for t in text:
                    lines.append(f'{tabs}{tab}{t}')
                lines.append(f'{tabs}</{self._tag}>')
        elif elements:
            lines.append(line + '>')
            for node in elements:
                node._to_xml(indent, level + 1, lines)
            lines.append(f'{tabs}</{self._tag}>')
        else:
            lines.append(line + '/>')


class Document(Node):
    """
    An XML document.
    :param root: The only root element.
    """

    def __init__(self, root: Element):
        self._root = root

    def add(self, *elements: 'Element') -> 'Document':
        self._root.add(*elements)
        return self

    def to_xml(self, indent: int = 2) -> str:
        xml = self._root.to_xml()
        return f'<?xml version="1.0" encoding="UTF-8"?>\n{xml}\n'



# The MIT License (MIT)
# Copyright (c) 2019 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import warnings
from typing import Union, Sequence, Optional, AbstractSet, Set

import numpy as np
import pyproj
import shapely.geometry
import shapely.ops
import shapely.wkt
import xarray as xr

from xcube.core.geom import GeometryLike
from xcube.core.geom import normalize_geometry
from xcube.core.geom import get_dataset_geometry
from xcube.core.geom import mask_dataset_by_geometry
from xcube.core.gridmapping import GridMapping
from xcube.core.select import select_variables_subset
from xcube.util.timeindex import ensure_time_index_compatible
from xcube.util.assertions import assert_instance
from xcube.constants import CRS_CRS84

Date = Union[np.datetime64, str]

AGG_MEAN = 'mean'
AGG_MEDIAN = 'median'
AGG_STD = 'std'
AGG_MIN = 'min'
AGG_MAX = 'max'
AGG_COUNT = 'count'

MUST_LOAD = True
CAN_COMPUTE = False

AGG_METHODS = {
    AGG_MEAN: CAN_COMPUTE,
    AGG_MEDIAN: MUST_LOAD,
    AGG_STD: CAN_COMPUTE,
    AGG_MIN: CAN_COMPUTE,
    AGG_MAX: CAN_COMPUTE,
    AGG_COUNT: CAN_COMPUTE
}


def get_time_series(
        cube: xr.Dataset,
        grid_mapping: Optional[GridMapping] = None,
        geometry: Optional[GeometryLike] = None,
        var_names: Optional[Sequence[str]] = None,
        start_date: Optional[Date] = None,
        end_date: Optional[Date] = None,
        agg_methods: Union[str, Sequence[str], AbstractSet[str]] = AGG_MEAN,
        use_groupby: bool = False,
        cube_asserted: Optional[bool] = None
) -> Optional[xr.Dataset]:
    """
    Get a time series dataset from a data *cube*.

    *geometry* may be provided as a (shapely) geometry object, a valid
    GeoJSON object, a valid WKT string,
    a sequence of box coordinates (x1, y1, x2, y2), or point coordinates
    (x, y). If *geometry* covers an area,
    i.e. is not a point, the function aggregates the variables to compute a
    mean value and if desired,
    the number of valid observations and the standard deviation.

    *start_date* and *end_date* may be provided as a numpy.datetime64 or an
    ISO datetime string.

    Returns a time-series dataset whose data variables have a time dimension
    but no longer have spatial dimensions,
    hence the resulting dataset's variables will only have N-2 dimensions.
    A global attribute ``max_number_of_observations`` will be set to the
    maximum number of observations
    that could have been made in each time step.
    If the given *geometry* does not overlap the cube's boundaries, or if not
    output variables remain,
    the function returns ``None``.

    :param cube: The xcube dataset
    :param grid_mapping: Grid mapping of *cube*.
    :param geometry: Optional geometry
    :param var_names: Optional sequence of names of variables to be included.
    :param start_date: Optional start date.
    :param end_date: Optional end date.
    :param agg_methods: Aggregation methods. May be single string or
        sequence of strings. Possible values are
        'mean', 'median', 'min', 'max', 'std', 'count'. Defaults to 'mean'.
        Ignored if geometry is a point.
    :param use_groupby: Use group-by operation. May increase or decrease
        runtime performance and/or memory consumption.
    :param cube_asserted: Deprecated and ignored since xcube 0.11.0.
        No replacement.
    """
    if cube_asserted is not None:
        warnings.warn('cube_asserted has been deprecated'
                      ' and will be removed soon.', DeprecationWarning)
    assert_instance(cube, xr.Dataset)
    if grid_mapping is not None:
        assert_instance(grid_mapping, GridMapping)
    else:
        grid_mapping = GridMapping.from_dataset(cube)

    geometry = normalize_geometry(geometry)
    if geometry is not None and not grid_mapping.crs.is_geographic:
        project = pyproj.Transformer.from_crs(CRS_CRS84,
                                              grid_mapping.crs,
                                              always_xy=True).transform
        geometry = shapely.ops.transform(project, geometry)

    dataset = select_variables_subset(cube, var_names)
    if len(dataset.data_vars) == 0:
        return None

    if start_date is not None or end_date is not None:
        date_slice = slice(start_date, end_date)
        safe_slice = ensure_time_index_compatible(dataset, date_slice)
        dataset = dataset.sel(time=safe_slice)

    x_name, y_name = grid_mapping.xy_dim_names
    if isinstance(geometry, shapely.geometry.Point):
        bounds = get_dataset_geometry(dataset)
        if not bounds.contains(geometry):
            return None
        indexers = {x_name: geometry.x, y_name: geometry.y}
        dataset = dataset.sel(**indexers, method='Nearest')
        return dataset.assign_attrs(max_number_of_observations=1)

    agg_methods = normalize_agg_methods(agg_methods)

    if geometry is not None:
        dataset = mask_dataset_by_geometry(dataset, geometry,
                                           save_geometry_mask='__mask__')
        if dataset is None:
            return None
        mask = dataset['__mask__']
        max_number_of_observations = np.count_nonzero(mask)
        dataset = dataset.drop_vars(['__mask__'])
    else:
        max_number_of_observations = \
            dataset[y_name].size * dataset[x_name].size

    must_load = len(agg_methods) > 1 \
                or any(AGG_METHODS[agg_method] == MUST_LOAD
                       for agg_method in agg_methods)
    if must_load:
        dataset.load()

    agg_datasets = []
    if use_groupby:
        time_group = dataset.groupby('time')
        for agg_method in agg_methods:
            method = getattr(time_group, agg_method)
            if agg_method == 'count':
                agg_dataset = method(dim=xr.ALL_DIMS)
            else:
                agg_dataset = method(dim=xr.ALL_DIMS, skipna=True)
            agg_datasets.append(agg_dataset)
    else:
        for agg_method in agg_methods:
            method = getattr(dataset, agg_method)
            if agg_method == 'count':
                agg_dataset = method(dim=(y_name, x_name))
            else:
                agg_dataset = method(dim=(y_name, x_name), skipna=True)
            agg_datasets.append(agg_dataset)

    agg_datasets = [
        agg_dataset.rename(
            name_dict={v: f"{v}_{agg_method}" for v in agg_dataset.data_vars}
        )
        for agg_method, agg_dataset in zip(agg_methods, agg_datasets)
    ]

    ts_dataset = xr.merge(agg_datasets)
    ts_dataset = ts_dataset.assign_attrs(
        max_number_of_observations=max_number_of_observations
    )

    return ts_dataset


def normalize_agg_methods(agg_methods: Union[str, Sequence[str]],
                          exception_type=ValueError) -> Set[str]:
    agg_methods = agg_methods or [AGG_MEAN]
    if isinstance(agg_methods, str):
        agg_methods = [agg_methods]
    agg_methods = set(agg_methods)
    invalid_agg_methods = agg_methods - set(AGG_METHODS.keys())
    if invalid_agg_methods:
        s = 's' if len(invalid_agg_methods) > 1 else ''
        raise exception_type(
            f'invalid aggregation method{s}:'
            f' {", ".join(sorted(list(invalid_agg_methods)))}'
        )
    return agg_methods



# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import math

import pyproj
import xarray as xr

from xcube.core.gridmapping import GridMapping
from xcube.core.select import select_spatial_subset
from xcube.core.select import select_temporal_subset
from xcube.core.select import select_variables_subset
from .transformer import CubeTransformer
from .transformer import TransformedCube
from ..config import CubeConfig


class CubeSubsetter(CubeTransformer):

    def transform_cube(self,
                       cube: xr.Dataset,
                       gm: GridMapping,
                       cube_config: CubeConfig) -> TransformedCube:

        desired_var_names = cube_config.variable_names
        if desired_var_names:
            cube = select_variables_subset(cube,
                                           var_names=desired_var_names)
            cube_config = cube_config.drop_props('variable_names')

        desired_bbox = cube_config.bbox
        if desired_bbox is not None:
            # Find out whether its possible to make a spatial subset
            # without resampling. First, grid mapping must be regular.
            can_do_spatial_subset = False
            if gm.is_regular:
                can_do_spatial_subset = True
                # Current spatial resolution must be the
                # desired spatial resolution, otherwise spatial resampling
                # is required later, which will include the desired
                # subsetting.
                desired_res = cube_config.spatial_res
                if desired_res is not None \
                        and not (math.isclose(gm.x_res, desired_res)
                                 and math.isclose(gm.y_res, desired_res)):
                    can_do_spatial_subset = False
                if can_do_spatial_subset:
                    # Finally, the desired CRS must be equal to the current
                    # one, or they must both be geographic.
                    desired_crs = cube_config.crs
                    if desired_crs:
                        desired_crs = pyproj.CRS.from_string(desired_crs)
                        if desired_crs != gm.crs \
                                and not (desired_crs.is_geographic
                                         and gm.crs.is_geographic):
                            can_do_spatial_subset = False
            if can_do_spatial_subset:
                cube = select_spatial_subset(cube,
                                             xy_bbox=desired_bbox)
                # Now that we have a new cube subset, we must adjust
                # its grid mapping.
                gm = GridMapping.from_dataset(
                    cube,
                    crs=gm.crs,
                )
                # Consume spatial properties
                cube_config = cube_config.drop_props(['bbox',
                                                      'spatial_res',
                                                      'crs'])

        desired_time_range = cube_config.time_range
        if desired_time_range:
            cube = select_temporal_subset(cube,
                                          time_range=desired_time_range)
            cube_config = cube_config.drop_props('time_range')

        return cube, gm, cube_config



# The MIT License (MIT)
# Copyright (c) 2023 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.
from typing import Optional

import pyproj
import xarray as xr

from xcube.core.gridmapping import GridMapping
from xcube.server.api import ApiError
from xcube.webapi.ows.coverages.util import get_h_dim, get_v_dim
from xcube.webapi.ows.coverages.request import CoverageRequest


class CoverageScaling:
    """Representation of a scaling applied to an OGC coverage

    This class represents the scaling specified in an OGC coverage request.
    It is instantiated using a `CoverageRequest` instance, and can apply
    itself to a `GridMapping` instance.
    """

    _scale: Optional[tuple[float, float]] = None
    _final_size: Optional[tuple[int, int]] = None
    _initial_size: tuple[int, int]
    _crs: pyproj.CRS
    _x_name: str
    _y_name: str

    def __init__(
        self, request: CoverageRequest, crs: pyproj.CRS, ds: xr.Dataset
    ):
        """Create a new scaling from a coverages request object

        :param request: a request optionally including scaling parameters.
          If any scaling parameters are present, the returned instance will
          correspond to them. If no scaling parameters are present, a
          default scaling will be used (currently 1:1)
        :param crs: the CRS of the dataset to be scaled
        :param ds: the dataset to be scaled
        """
        h_dim = get_h_dim(ds)
        v_dim = get_v_dim(ds)
        for d in h_dim, v_dim:
            size = ds.dims[d]
            if size == 0:
                # Requirement 8C currently specifies a 204 rather than 404 here,
                # but spec will soon be updated to allow 404 as an alternative.
                # (J. Jacovella-St-Louis, pers. comm., 2023-11-27).
                raise ApiError.NotFound(
                    f'Requested coverage contains no data: {d} has zero size.'
                )
        self._initial_size = ds.dims[h_dim], ds.dims[v_dim]

        self._crs = crs
        self._y_name = self.get_axis_from_crs(
            {'lat', 'latitude', 'geodetic latitude', 'n', 'north', 'y'}
        )
        self._x_name = self.get_axis_from_crs(
            {'longitude', 'geodetic longitude', 'lon', 'e', 'east', 'x'}
        )

        # The standard doesn't define behaviour when multiple scale
        # parameters are given. We choose to handle one and ignore the
        # others in such cases.
        if request.scale_factor is not None:
            self._scale = request.scale_factor, request.scale_factor
        elif request.scale_axes is not None:
            self._scale = self._get_xy_values(request.scale_axes)
        elif request.scale_size is not None:
            # The standard allows floats for "scale-size" but mandates:
            # "The returned coverage SHALL contain exactly the specified number
            # of sample values". We can't return a fractional number of sample
            # values, so truncate to int here.
            x, y = self._get_xy_values(request.scale_size)
            self._final_size = int(x), int(y)
        else:
            # The standard doesn't mandate this as a default; in future, we
            # might choose to downscale automatically if a large coverage
            # is requested without an explicit scaling parameter.
            self._scale = (1, 1)

    @property
    def factor(self) -> tuple[float, float]:
        """Return the two-dimensional scale factor of this scaling

        The components of the scale tuple are expressed as downscaling
        factors: values greater than 1 imply that the rescaled size
        of the coverage in the corresponding dimension will be smaller than
        the original size, and vice versa.

        If the scaling was initially specified as a final size rather than
        a factor, the factor property is an estimate based on the dataset
        dimensions; the effective factor may be different when the scaling
        is applied to a GridMapping.

        :return: a 2-tuple of the x and y scale factors, in that order
        """
        if self._scale is not None:
            return self._scale
        else:
            x_initial, y_initial = self._initial_size
            x_final, y_final = self._final_size
            return x_initial / x_final, y_initial / y_final

    @property
    def size(self) -> tuple[float, float]:
        """Return the final coverage size produced by this scaling

        :return: a 2-tuple of the scaled x and y sizes, in that order,
          in pixels
        """
        if self._final_size is not None:
            return self._final_size
        else:
            x_initial, y_initial = self._initial_size
            x_scale, y_scale = self._scale
            return x_initial / x_scale, y_initial / y_scale

    def _get_xy_values(
        self, axis_to_value: dict[str, float]
    ) -> tuple[float, float]:
        x, y = None, None
        for axis in axis_to_value:
            if axis.lower()[:3] in ['x', 'e', 'eas', 'lon', self._x_name]:
                x = axis_to_value[axis]
            if axis.lower()[:3] in ['y', 'n', 'nor', 'lat', self._y_name]:
                y = axis_to_value[axis]
        return x, y

    def get_axis_from_crs(self, valid_identifiers: set[str]) -> Optional[str]:
        """Find an axis abbreviation via a set of possible axis identifiers

        This method operates on the CRS with which this scaling was
        instantiated. It returns the abbreviation of the first axis in
        the CRS which has either a name or abbreviation matching any string
        in the supplied set.

        :param valid_identifiers: a set of axis identifiers
        :return: the abbreviation of the first axis in this scaling's CRS
          whose name or abbreviation is in the supplied set, or `None` if
          no such axis exists
        """
        for axis in self._crs.axis_info:
            if not hasattr(axis, 'abbrev'):
                continue
            identifiers = set(
                map(
                    lambda attr: getattr(axis, attr, '').lower(),
                    ['name', 'abbrev'],
                )
            )
            if identifiers & valid_identifiers:
                return axis.abbrev
        return None

    def apply(self, gm: GridMapping) -> GridMapping:
        """Apply this scaling to a grid mapping

        The supplied grid mapping is regularized before being scaled.

        :param gm: a grid mapping to be scaled
        :return: the supplied grid mapping, scaled according to this scaling.
          If this scaling is 1:1, the returned grid mapping may be the
          original object.
        """
        if self.factor == (1, 1):
            return gm
        else:
            regular = gm.to_regular()
            source = regular.size
            # Even if the scaling was specified as a factor, we calculate
            # from the (inferred) final size. If a factor was given,
            # self.size is the final size as calculated from the originally
            # specified dataset, which is what the client would expect. The
            # regularized GridMapping might have a different size,
            # so we don't want to apply a specified factor to it directly.
            return regular.scale(
                (self.size[0] / source[0], self.size[1] / source[1])
            )



# The MIT License (MIT)
# Copyright (c) 2023 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import math
from typing import Union, Callable, Optional, \
    Sequence, Tuple, Mapping, Hashable, Any

import numpy as np
import xarray as xr
from dask import array as da
from dask_image import ndinterp

from xcube.core.gridmapping import GridMapping
from xcube.core.gridmapping.helpers import AffineTransformMatrix
from xcube.util.assertions import assert_true
from .cf import maybe_encode_grid_mapping

NDImage = Union[np.ndarray, da.Array]
Aggregator = Callable[[NDImage], NDImage]


def affine_transform_dataset(
        dataset: xr.Dataset,
        source_gm: GridMapping,
        target_gm: GridMapping,
        var_configs: Mapping[Hashable, Mapping[str, Any]] = None,
        encode_cf: bool = True,
        gm_name: Optional[str] = None,
        reuse_coords: bool = False,
) -> xr.Dataset:
    """
    Resample dataset according to an affine transformation.

    The affine transformation will be applied only if the CRS of
    *source_gm* and the CRS of *target_gm* are both geographic or equal.
    Otherwise, a ``ValueError`` will be raised.

    :param dataset: The source dataset
    :param source_gm: Source grid mapping of *dataset*.
        Must be regular. Must have same CRS as *target_gm*.
    :param target_gm: Target grid mapping.
        Must be regular. Must have same CRS as *source_gm*.
    :param var_configs: Optional resampling configurations
        for individual variables.
    :param encode_cf: Whether to encode the target grid mapping
        into the resampled dataset in a CF-compliant way.
        Defaults to ``True``.
    :param gm_name: Name for the grid mapping variable.
        Defaults to "crs". Used only if *encode_cf* is ``True``.
    :param reuse_coords: Whether to either reuse target
        coordinate arrays from target_gm or to compute
        new ones.
    :return: The resampled target dataset.
    """
    # Are source and target both geographic grid mappings?
    both_geographic = source_gm.crs.is_geographic \
                      and target_gm.crs.is_geographic
    if not (both_geographic or source_gm.crs == target_gm.crs):
        raise ValueError(f'CRS of source_gm and target_gm must be equal,'
                         f' was "{source_gm.crs.name}"'
                         f' and "{target_gm.crs.name}"')
    GridMapping.assert_regular(source_gm, name='source_gm')
    GridMapping.assert_regular(target_gm, name='target_gm')
    resampled_dataset = resample_dataset(
        dataset=dataset,
        matrix=target_gm.ij_transform_to(source_gm),
        size=target_gm.size,
        tile_size=target_gm.tile_size,
        xy_dim_names=source_gm.xy_dim_names,
        var_configs=var_configs
    )
    has_bounds = any(dataset[var_name].attrs.get('bounds')
                     for var_name in source_gm.xy_var_names)
    new_coords = target_gm.to_coords(
        xy_var_names=source_gm.xy_var_names,
        xy_dim_names=source_gm.xy_dim_names,
        exclude_bounds=not has_bounds,
        reuse_coords=reuse_coords
    )
    return maybe_encode_grid_mapping(
        encode_cf,
        resampled_dataset.assign_coords(new_coords),
        target_gm,
        gm_name
    )


def resample_dataset(
        dataset: xr.Dataset,
        matrix: AffineTransformMatrix,
        size: Tuple[int, int],
        tile_size: Tuple[int, int],
        xy_dim_names: Tuple[str, str],
        var_configs: Mapping[Hashable, Mapping[str, Any]] = None,
) -> xr.Dataset:
    """
    Resample dataset according to an affine transformation.

    :param dataset: The source dataset
    :param matrix: Affine transformation matrix.
    :param size: Target image size.
    :param tile_size: Target image tile size.
    :param xy_dim_names: Names of the spatial dimensions.
    :param var_configs: Optional resampling configurations
        for individual variables.
    :return: The resampled target dataset.
    """
    ((i_scale, _, i_off), (_, j_scale, j_off)) = matrix
    width, height = size
    tile_width, tile_height = tile_size
    x_dim, y_dim = xy_dim_names
    yx_dims = (y_dim, x_dim)
    coords = dict()
    var_configs = var_configs or {}
    data_vars = dict()
    for k, var in dataset.variables.items():
        new_var = None
        if var.ndim >= 2 and var.dims[-2:] == yx_dims:
            var_config = var_configs.get(k, dict())
            if np.issubdtype(var.dtype, np.integer) \
                    or np.issubdtype(var.dtype, bool):
                spline_order = 0
                aggregator = None
                recover_nan = False
            else:
                spline_order = 1
                aggregator = np.nanmean
                recover_nan = True
            var_data = resample_ndimage(
                var.data,
                scale=(j_scale, i_scale),
                offset=(j_off, i_off),
                shape=(height, width),
                chunks=(tile_height, tile_width),
                spline_order=var_config.get('spline_order', spline_order),
                aggregator=var_config.get('aggregator', aggregator),
                recover_nan=var_config.get('recover_nan', recover_nan),
            )
            new_var = xr.DataArray(var_data,
                                   dims=var.dims,
                                   attrs=var.attrs)
        elif x_dim not in var.dims and y_dim not in var.dims:
            new_var = var.copy()
        if new_var is not None:
            if k in dataset.coords:
                coords[k] = new_var
            elif k in dataset.data_vars:
                data_vars[k] = new_var

    return xr.Dataset(data_vars=data_vars,
                      coords=coords,
                      attrs=dataset.attrs)


def resample_ndimage(
        image: NDImage,
        scale: Union[float, Tuple[float, float]] = 1,
        offset: Union[float, Tuple[float, float]] = None,
        shape: Union[int, Tuple[int, int]] = None,
        chunks: Sequence[int] = None,
        spline_order: int = 1,
        aggregator: Optional[Aggregator] = np.nanmean,
        recover_nan: bool = False
) -> da.Array:
    image = da.asarray(image)
    offset = _normalize_offset(offset, image.ndim)
    scale = _normalize_scale(scale, image.ndim)
    if shape is None:
        shape = resize_shape(image.shape, scale)
    else:
        shape = _normalize_shape(shape, image)
    chunks = _normalize_chunks(chunks, shape)
    scale_y, scale_x = scale[-2], scale[-1]
    divisor_x = math.ceil(abs(scale_x))
    divisor_y = math.ceil(abs(scale_y))
    if (divisor_x >= 2 or divisor_y >= 2) and aggregator is not None:
        # Downsampling
        # ------------
        axes = {image.ndim - 2: divisor_y, image.ndim - 1: divisor_x}
        elongation = _normalize_scale((scale_y / divisor_y,
                                       scale_x / divisor_x), image.ndim)
        larger_shape = resize_shape(shape, (divisor_y, divisor_x),
                                    divisor_x=divisor_x,
                                    divisor_y=divisor_y)
        # print('Downsampling: ', scale)
        # print('  divisor:', (divisor_y, divisor_x))
        # print('  elongation:', elongation)
        # print('  shape:', shape)
        # print('  larger_shape:', larger_shape)
        divisible_chunks = _make_divisible_tiles(larger_shape,
                                                 divisor_x, divisor_y)
        image = _transform_array(image,
                                 elongation, offset,
                                 larger_shape, divisible_chunks,
                                 spline_order, recover_nan)
        image = da.coarsen(aggregator, image, axes)
        if shape != image.shape:
            image = image[..., 0:shape[-2], 0:shape[-1]]
        if chunks is not None:
            image = image.rechunk(chunks)
    else:
        # Upsampling
        # ----------
        # print('Upsampling: ', scale)
        image = _transform_array(image,
                                 scale, offset,
                                 shape, chunks,
                                 spline_order, recover_nan)
    return image


def _transform_array(image: da.Array,
                     scale: Tuple[float, ...],
                     offset: Tuple[float, ...],
                     shape: Tuple[int, ...],
                     chunks: Optional[Tuple[int, ...]],
                     spline_order: int,
                     recover_nan: bool) -> da.Array:
    """
    Apply affine transformation to ND-image.

    :param image: ND-image with shape (..., size_y, size_x)
    :param scale: Scaling factors (1, ..., 1, sy, sx)
    :param offset: Offset values (0, ..., 0, oy, ox)
    :param shape: (..., size_y, size_x)
    :param chunks: (..., chunk_size_y, chunk_size_x)
    :param spline_order: 0 ... 5
    :param recover_nan: True/False
    :return: Transformed ND-image.
    """
    assert_true(len(scale) == image.ndim, 'invalid scale')
    assert_true(len(offset) == image.ndim, 'invalid offset')
    assert_true(len(shape) == image.ndim, 'invalid shape')
    assert_true(chunks is None or len(chunks) == image.ndim,
                'invalid chunks')
    if _is_no_op(image, scale, offset, shape):
        return image
    # As of scipy 0.18, matrix = scale is no longer supported.
    # Therefore we use the diagonal matrix form here,
    # where scale is the diagonal.
    matrix = np.diag(scale)
    at_kwargs = dict(
        offset=offset,
        order=spline_order,
        output_shape=shape,
        output_chunks=chunks,
        mode='constant',
    )
    if recover_nan and spline_order > 0:
        # We can "recover" values that are neighbours to NaN values
        # that would otherwise become NaN too.
        mask = da.isnan(image)
        # First check if there are NaN values ar all
        if da.any(mask):
            # Yes, then
            # 1. replace NaN by zero
            filled_im = da.where(mask, 0.0, image)
            # 2. transform the zeo-filled image
            scaled_im = ndinterp.affine_transform(filled_im,
                                                  matrix,
                                                  **at_kwargs,
                                                  cval=0.0)
            # 3. transform the inverted mask
            scaled_norm = ndinterp.affine_transform(1.0 - mask,
                                                    matrix,
                                                    **at_kwargs,
                                                    cval=0.0)
            # 4. put back NaN where there was zero,
            #    otherwise decode using scaled mask
            return da.where(da.isclose(scaled_norm, 0.0),
                            np.nan, scaled_im / scaled_norm)

    # No dealing with NaN required
    return ndinterp.affine_transform(image, matrix, **at_kwargs, cval=np.nan)


def resize_shape(shape: Sequence[int],
                 scale: Union[float, Tuple[float, ...]],
                 divisor_x: int = 1,
                 divisor_y: int = 1) -> Tuple[int, ...]:
    scale = _normalize_scale(scale, len(shape))
    height, width = shape[-2], shape[-1]
    scale_y, scale_x = scale[-2], scale[-1]
    wf = width * abs(scale_x)
    hf = height * abs(scale_y)
    w = divisor_x * math.ceil(wf / divisor_x)
    h = divisor_y * math.ceil(hf / divisor_y)
    return tuple(shape[0:-2]) + (h, w)


def _make_divisible_tiles(larger_shape: Tuple[int, ...],
                          divisor_x: int,
                          divisor_y: int) -> Tuple[int, ...]:
    w = min(larger_shape[-1],
            divisor_x * ((2048 + divisor_x - 1) // divisor_x))
    h = min(larger_shape[-2],
            divisor_y * ((2048 + divisor_y - 1) // divisor_y))
    return (len(larger_shape) - 2) * (1,) + (h, w)


def _normalize_image(im: NDImage) -> da.Array:
    return da.asarray(im)


def _normalize_offset(offset: Optional[Sequence[float]],
                      ndim: int) -> Tuple[int, ...]:
    return _normalize_pair(offset, 0.0, ndim, 'offset')


def _normalize_scale(scale: Optional[Sequence[float]],
                     ndim: int) -> Tuple[int, ...]:
    return _normalize_pair(scale, 1.0, ndim, 'scale')


def _normalize_pair(pair: Optional[Sequence[float]],
                    default: float,
                    ndim: int,
                    name: str) -> Tuple[int, ...]:
    if pair is None:
        pair = [default, default]
    elif isinstance(pair, (int, float)):
        pair = [pair, pair]
    elif len(pair) != 2:
        raise ValueError(f'illegal image {name}')
    return (ndim - 2) * (default,) + tuple(pair)


def _normalize_shape(shape: Optional[Sequence[int]],
                     im: NDImage) -> Tuple[int, ...]:
    if shape is None:
        return im.shape
    if len(shape) != 2:
        raise ValueError('illegal image shape')
    return im.shape[0:-2] + tuple(shape)


def _normalize_chunks(chunks: Optional[Sequence[int]],
                      shape: Tuple[int, ...]) -> Optional[Tuple[int, ...]]:
    if chunks is None:
        return None
    if len(chunks) < 2 or len(chunks) > len(shape):
        raise ValueError('illegal image chunks')
    return (len(shape) - len(chunks)) * (1,) + tuple(chunks)


def _is_no_op(im: NDImage,
              scale: Sequence[float],
              offset: Sequence[float],
              shape: Tuple[int, ...]):
    return shape == im.shape \
           and all(math.isclose(s, 1) for s in scale) \
           and all(math.isclose(o, 0) for o in offset)



# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import abc
import math
from typing import Tuple, Union, Dict

import dask.array as da
import numpy as np
import pyproj
import xarray as xr

from xcube.util.assertions import assert_instance
from xcube.util.assertions import assert_true
from .base import DEFAULT_TOLERANCE
from .base import GridMapping
from .helpers import _assert_valid_xy_names
from .helpers import _default_xy_var_names
from .helpers import _normalize_crs
from .helpers import _normalize_int_pair
from .helpers import _to_int_or_float
from .helpers import from_lon_360
from .helpers import round_to_fraction
from .helpers import to_lon_360

_ER = 6371000


class CoordsGridMapping(GridMapping, abc.ABC):
    """
    Grid mapping constructed from 1D/2D coordinate
    variables and a CRS.
    """

    @property
    def x_coords(self):
        assert isinstance(self._x_coords, xr.DataArray)
        return self._x_coords

    @property
    def y_coords(self):
        assert isinstance(self._y_coords, xr.DataArray)
        return self._y_coords

    def _new_x_coords(self) -> xr.DataArray:
        # Should never come here
        return self._x_coords

    def _new_y_coords(self) -> xr.DataArray:
        # Should never come here
        return self._y_coords


class Coords1DGridMapping(CoordsGridMapping):
    """
    Grid mapping constructed from
    1D coordinate variables and a CRS.
    """

    def _new_xy_coords(self) -> xr.DataArray:
        y, x = xr.broadcast(self._y_coords, self._x_coords)
        return xr.concat([x, y], dim='coord') \
            .chunk(self.xy_coords_chunks)


class Coords2DGridMapping(CoordsGridMapping):
    """
    Grid mapping constructed from
    2D coordinate variables and a CRS.
    """

    def _new_xy_coords(self) -> xr.DataArray:
        return xr.concat([self._x_coords, self._y_coords], dim='coord') \
            .chunk(self.xy_coords_chunks)


def new_grid_mapping_from_coords(
        x_coords: xr.DataArray,
        y_coords: xr.DataArray,
        crs: Union[str, pyproj.crs.CRS],
        *,
        tile_size: Union[int, Tuple[int, int]] = None,
        tolerance: float = DEFAULT_TOLERANCE,
) -> GridMapping:
    crs = _normalize_crs(crs)
    assert_instance(x_coords, xr.DataArray, name='x_coords')
    assert_instance(y_coords, xr.DataArray, name='y_coords')
    assert_true(x_coords.ndim in (1, 2),
                'x_coords and y_coords must be either 1D or 2D arrays')
    assert_instance(tolerance, float, name='tolerance')
    assert_true(tolerance > 0.0,
                'tolerance must be greater zero')

    if x_coords.name and y_coords.name:
        xy_var_names = str(x_coords.name), str(y_coords.name)
    else:
        xy_var_names = _default_xy_var_names(crs)

    tile_size = _normalize_int_pair(tile_size, default=None)
    is_lon_360 = None  # None means "not yet known"
    if crs.is_geographic:
        is_lon_360 = bool(np.any(x_coords > 180))

    x_res = 0
    y_res = 0

    if x_coords.ndim == 1:
        # We have 1D x,y coordinates
        cls = Coords1DGridMapping

        assert_true(x_coords.size >= 2 and y_coords.size >= 2,
                    'sizes of x_coords and y_coords 1D arrays must be >= 2')

        size = x_coords.size, y_coords.size

        x_dim, y_dim = x_coords.dims[0], y_coords.dims[0]

        x_diff = _abs_no_zero(x_coords.diff(dim=x_dim).values)
        y_diff = _abs_no_zero(y_coords.diff(dim=y_dim).values)

        if not is_lon_360 and crs.is_geographic:
            is_anti_meridian_crossed = np.any(np.nanmax(x_diff) > 180)
            if is_anti_meridian_crossed:
                x_coords = to_lon_360(x_coords)
                x_diff = _abs_no_zero(x_coords.diff(dim=x_dim))
                is_lon_360 = True

        x_res, y_res = x_diff[0], y_diff[0]
        x_diff_equal = np.allclose(x_diff, x_res, atol=tolerance)
        y_diff_equal = np.allclose(y_diff, y_res, atol=tolerance)
        is_regular = x_diff_equal and y_diff_equal
        if is_regular:
            x_res = round_to_fraction(x_res, 5, 0.25)
            y_res = round_to_fraction(y_res, 5, 0.25)
        else:
            x_res = round_to_fraction(float(np.nanmedian(x_diff)), 2, 0.5)
            y_res = round_to_fraction(float(np.nanmedian(y_diff)), 2, 0.5)

        if tile_size is None \
                and x_coords.chunks is not None \
                and y_coords.chunks is not None:
            tile_size = (max(0, *x_coords.chunks[0]),
                         max(0, *y_coords.chunks[0]))

        # Guess j axis direction
        is_j_axis_up = bool(y_coords[0] < y_coords[-1])

    else:
        # We have 2D x,y coordinates
        cls = Coords2DGridMapping

        assert_true(x_coords.shape == y_coords.shape,
                    'shapes of x_coords and y_coords'
                    ' 2D arrays must be equal')
        assert_true(x_coords.dims == y_coords.dims,
                    'dimensions of x_coords and y_coords'
                    ' 2D arrays must be equal')

        y_dim, x_dim = x_coords.dims

        height, width = x_coords.shape
        size = width, height

        x = da.asarray(x_coords)
        y = da.asarray(y_coords)

        x_x_diff = _abs_no_nan(da.diff(x, axis=1))
        x_y_diff = _abs_no_nan(da.diff(x, axis=0))
        y_x_diff = _abs_no_nan(da.diff(y, axis=1))
        y_y_diff = _abs_no_nan(da.diff(y, axis=0))

        if not is_lon_360 and crs.is_geographic:
            is_anti_meridian_crossed = da.any(da.max(x_x_diff) > 180) \
                                       or da.any(da.max(x_y_diff) > 180)
            if is_anti_meridian_crossed:
                x_coords = to_lon_360(x_coords)
                x = da.asarray(x_coords)
                x_x_diff = _abs_no_nan(da.diff(x, axis=1))
                x_y_diff = _abs_no_nan(da.diff(x, axis=0))
                is_lon_360 = True

        is_regular = False

        if da.all(x_y_diff == 0) and da.all(y_x_diff == 0):
            x_res = x_x_diff[0, 0]
            y_res = y_y_diff[0, 0]
            is_regular = \
                da.allclose(x_x_diff[0, :], x_res, atol=tolerance) \
                and da.allclose(x_x_diff[-1, :], x_res, atol=tolerance) \
                and da.allclose(y_y_diff[:, 0], y_res, atol=tolerance) \
                and da.allclose(y_y_diff[:, -1], y_res, atol=tolerance)

        if not is_regular:
            # Let diff arrays have same shape as original by
            # doubling last rows and columns.
            x_x_diff_c = da.concatenate([x_x_diff, x_x_diff[:, -1:]], axis=1)
            y_x_diff_c = da.concatenate([y_x_diff, y_x_diff[:, -1:]], axis=1)
            x_y_diff_c = da.concatenate([x_y_diff, x_y_diff[-1:, :]], axis=0)
            y_y_diff_c = da.concatenate([y_y_diff, y_y_diff[-1:, :]], axis=0)
            # Find resolution via area
            x_abs_diff = da.sqrt(da.square(x_x_diff_c) + da.square(x_y_diff_c))
            y_abs_diff = da.sqrt(da.square(y_x_diff_c) + da.square(y_y_diff_c))
            if crs.is_geographic:
                # Convert degrees into meters
                x_abs_diff_r = da.radians(x_abs_diff)
                y_abs_diff_r = da.radians(y_abs_diff)
                x_abs_diff = _ER * da.cos(x_abs_diff_r) * y_abs_diff_r
                y_abs_diff = _ER * y_abs_diff_r
            xy_areas = (x_abs_diff * y_abs_diff).flatten()
            xy_areas = da.where(xy_areas > 0, xy_areas, np.nan)
            # Get indices of min and max area
            xy_area_index_min = da.nanargmin(xy_areas)
            xy_area_index_max = da.nanargmax(xy_areas)
            # Convert area to edge length
            xy_res_min = math.sqrt(xy_areas[xy_area_index_min])
            xy_res_max = math.sqrt(xy_areas[xy_area_index_max])
            # Empirically weight min more than max
            xy_res = 0.7 * xy_res_min + 0.3 * xy_res_max
            if crs.is_geographic:
                # Convert meters back into degrees
                # print(f'xy_res in meters: {xy_res}')
                xy_res = math.degrees(xy_res / _ER)
                # print(f'xy_res in degrees: {xy_res}')
            # Because this is an estimation, we can round to a nice number
            xy_res = round_to_fraction(xy_res, digits=1, resolution=0.5)
            x_res, y_res = float(xy_res), float(xy_res)

        if tile_size is None and x_coords.chunks is not None:
            j_chunks, i_chunks = x_coords.chunks
            tile_size = max(0, *i_chunks), max(0, *j_chunks)

        if tile_size is not None:
            tile_width, tile_height = tile_size
            x_coords = x_coords.chunk((tile_height, tile_width))
            y_coords = y_coords.chunk((tile_height, tile_width))

        # Guess j axis direction
        is_j_axis_up = np.all(y_coords[0, :] < y_coords[-1, :]) or None

    assert_true(x_res > 0 and y_res > 0,
                'internal error: x_res and y_res could not be determined',
                exception_type=RuntimeError)

    x_res, y_res = _to_int_or_float(x_res), _to_int_or_float(y_res)
    x_res_05, y_res_05 = x_res / 2, y_res / 2
    x_min = _to_int_or_float(x_coords.min() - x_res_05)
    y_min = _to_int_or_float(y_coords.min() - y_res_05)
    x_max = _to_int_or_float(x_coords.max() + x_res_05)
    y_max = _to_int_or_float(y_coords.max() + y_res_05)

    if cls is Coords1DGridMapping and is_regular:
        from .regular import RegularGridMapping
        cls = RegularGridMapping

    return cls(x_coords=x_coords,
               y_coords=y_coords,
               crs=crs,
               size=size,
               tile_size=tile_size,
               xy_bbox=(x_min, y_min, x_max, y_max),
               xy_res=(x_res, y_res),
               xy_var_names=xy_var_names,
               xy_dim_names=(str(x_dim), str(y_dim)),
               is_regular=is_regular,
               is_lon_360=is_lon_360,
               is_j_axis_up=is_j_axis_up)


def _abs_no_zero(array: Union[xr.DataArray, da.Array, np.ndarray]):
    array = np.fabs(array)
    return np.where(np.isclose(array, 0),
                    np.nan, array)


def _abs_no_nan(array: Union[xr.DataArray, da.Array, np.ndarray]):
    array = np.fabs(array)
    return np.where(np.logical_or(np.isnan(array),
                                  np.isclose(array, 0)),
                    0, array)


def grid_mapping_to_coords(
        grid_mapping: GridMapping,
        xy_var_names: Tuple[str, str] = None,
        xy_dim_names: Tuple[str, str] = None,
        reuse_coords: bool = False,
        exclude_bounds: bool = False,
) -> Dict[str, xr.DataArray]:
    """
    Get CF-compliant axis coordinate variables and cell
    boundary coordinate variables.

    Defined only for grid mappings with regular x,y coordinates.

    :param grid_mapping: A regular grid mapping.
    :param xy_var_names: Optional coordinate variable
        names (x_var_name, y_var_name).
    :param xy_dim_names: Optional coordinate dimensions
        names (x_dim_name, y_dim_name).
    :param reuse_coords: Whether to either reuse target
        coordinate arrays from target_gm or to compute
        new ones.
    :param exclude_bounds: If True, do not create
        bounds coordinates. Defaults to False. Ignored if
        *reuse_coords* is True.
    :return: dictionary with coordinate variables
    """

    if xy_var_names:
        _assert_valid_xy_names(xy_var_names, name='xy_var_names')
    if xy_dim_names:
        _assert_valid_xy_names(xy_dim_names, name='xy_dim_names')

    if reuse_coords:
        try:
            # noinspection PyUnresolvedReferences
            x, y = grid_mapping.x_coords, grid_mapping.y_coords
        except AttributeError:
            x, y = None, None
        if isinstance(x, xr.DataArray) \
                and isinstance(y, xr.DataArray) \
                and x.ndim == 1 \
                and y.ndim == 1 \
                and x.size == grid_mapping.width \
                and y.size == grid_mapping.height:
            return {
                name: xr.DataArray(coord.values,
                                   dims=dim,
                                   attrs=coord.attrs)
                for name, dim, coord in zip(xy_var_names,
                                            xy_dim_names,
                                            (x, y))
            }

    x_name, y_name = xy_var_names or grid_mapping.xy_var_names
    x_dim_name, y_dim_name = xy_dim_names or grid_mapping.xy_dim_names
    w, h = grid_mapping.size
    x1, y1, x2, y2 = grid_mapping.xy_bbox
    x_res, y_res = grid_mapping.xy_res
    x_res_05 = x_res / 2
    y_res_05 = y_res / 2

    dtype = np.float64

    x_data = np.linspace(x1 + x_res_05, x2 - x_res_05, w, dtype=dtype)
    if grid_mapping.is_lon_360:
        x_data = from_lon_360(x_data)

    if grid_mapping.is_j_axis_up:
        y_data = np.linspace(y1 + y_res_05, y2 - y_res_05, h, dtype=dtype)
    else:
        y_data = np.linspace(y2 - y_res_05, y1 + y_res_05, h, dtype=dtype)

    if grid_mapping.crs.is_geographic:
        x_attrs = dict(
            long_name='longitude coordinate',
            standard_name='longitude',
            units='degrees_east'
        )
        y_attrs = dict(
            long_name='latitude coordinate',
            standard_name='latitude',
            units='degrees_north'
        )
    else:
        x_attrs = dict(
            long_name="x coordinate of projection",
            standard_name="projection_x_coordinate"
        )
        y_attrs = dict(
            long_name="y coordinate of projection",
            standard_name="projection_y_coordinate"
        )

    x_coords = xr.DataArray(x_data, dims=x_dim_name, attrs=x_attrs)
    y_coords = xr.DataArray(y_data, dims=y_dim_name, attrs=y_attrs)
    coords = {
        x_name: x_coords,
        y_name: y_coords,
    }
    if not exclude_bounds:
        x_bnds_0_data = np.linspace(x1, x2 - x_res, w, dtype=dtype)
        x_bnds_1_data = np.linspace(x1 + x_res, x2, w, dtype=dtype)

        if grid_mapping.is_lon_360:
            x_bnds_0_data = from_lon_360(x_bnds_0_data)
            x_bnds_1_data = from_lon_360(x_bnds_1_data)

        if grid_mapping.is_j_axis_up:
            y_bnds_0_data = np.linspace(y1, y2 - y_res, h, dtype=dtype)
            y_bnds_1_data = np.linspace(y1 + y_res, y2, h, dtype=dtype)
        else:
            y_bnds_0_data = np.linspace(y2, y1 + y_res, h, dtype=dtype)
            y_bnds_1_data = np.linspace(y2 - y_res, y1, h, dtype=dtype)

        bnds_dim_name = 'bnds'
        x_bnds_name = f'{x_name}_{bnds_dim_name}'
        y_bnds_name = f'{y_name}_{bnds_dim_name}'
        # Note, according to CF, bounds variables are not required to have
        # any attributes, so we don't pass any.
        x_bnds_coords = xr.DataArray(list(zip(x_bnds_0_data, x_bnds_1_data)),
                                     dims=[x_dim_name, bnds_dim_name])
        y_bnds_coords = xr.DataArray(list(zip(y_bnds_0_data, y_bnds_1_data)),
                                     dims=[y_dim_name, bnds_dim_name])
        x_coords.attrs.update(bounds=x_bnds_name)
        y_coords.attrs.update(bounds=y_bnds_name)
        coords.update({
            x_bnds_name: x_bnds_coords,
            y_bnds_name: y_bnds_coords,
        })

    return coords
